# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
# Author: Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import matplotlib.pyplot as plt
import numpy as np

import nifty8 as ift


def polynomial(coefficients, sampling_points):
    """Computes values of polynomial whose coefficients are stored in
    coefficients at sampling points. This is a quick version of the
    PolynomialResponse.

    Parameters
    ----------
    coefficients: Field
    sampling_points: Numpy array
    """

    if not (isinstance(coefficients, ift.Field)
            and isinstance(sampling_points, np.ndarray)):
        raise TypeError
    params = coefficients.val
    out = np.zeros_like(sampling_points)
    for ii in range(len(params)):
        out += params[ii] * sampling_points**ii
    return out


class PolynomialResponse(ift.LinearOperator):
    """Calculates values of a polynomial parameterized by input at sampling
    points.

    Parameters
    ----------
    domain: UnstructuredDomain
        The domain on which the coefficients of the polynomial are defined.
    sampling_points: Numpy array
        x-values of the sampling points.
    """

    def __init__(self, domain, sampling_points):
        if not (isinstance(domain, ift.UnstructuredDomain)
                and isinstance(sampling_points, np.ndarray)):
            raise TypeError
        self._domain = ift.DomainTuple.make(domain)
        tgt = ift.UnstructuredDomain(sampling_points.shape)
        self._target = ift.DomainTuple.make(tgt)
        self._capability = self.TIMES | self.ADJOINT_TIMES
        self._mat = np.empty((self.target.size, domain.size))
        for d in range(domain.size):
            self._mat.T[d] = sampling_points**d

    def apply(self, x, mode):
        self._check_input(x, mode)
        m = self._mat
        f = m if mode == self.TIMES else m.conj().T
        return ift.makeField(self._tgt(mode), f.dot(x.val))


def mock_fct(x):
    return np.sin(x**2 / 10) * x**3


def main():
    # Generate some mock data
    N_params = 10
    N_samples = 100
    size = (12,)

    x = ift.random.current_rng().random(size) * 10
    y = mock_fct(x)

    var = np.full_like(y, y.var() / 10)
    var[-2] *= 4
    var[5] /= 2
    var[5] -= 1

    # Set up minimization problem
    p_space = ift.UnstructuredDomain(N_params)
    params = ift.full(p_space, 0.)
    R = PolynomialResponse(p_space, x)
    ift.extra.check_linear_operator(R)

    d_space = R.target
    d = ift.makeField(d_space, y)
    N = ift.makeOp(ift.makeField(d_space, var), sampling_dtype=float)

    IC = ift.DeltaEnergyController(tol_rel_deltaE=1e-12, iteration_limit=200)
    likelihood_energy = ift.GaussianEnergy(d, N.inverse) @ R
    Ham = ift.StandardHamiltonian(likelihood_energy, IC)
    H = ift.EnergyAdapter(params, Ham, want_metric=True)

    # Minimize KL
    minimizer = ift.NewtonCG(IC)
    H, _ = minimizer(H)

    # Draw posterior samples
    metric = Ham(ift.Linearization.make_var(H.position, want_metric=True)).metric
    samples = [metric.draw_sample(from_inverse=True) + H.position
               for _ in range(N_samples)]

    # Plotting
    plt.errorbar(x, y, np.sqrt(var), fmt='ko', label='Data with error bars')
    xmin, xmax = x.min(), x.max()
    xs = np.linspace(xmin, xmax, 100)

    # Ground truth
    y_true = mock_fct(xs)

    sc = ift.StatCalculator()
    for ii in range(len(samples)):
        sc.add(samples[ii])
        ys = polynomial(samples[ii], xs)
        if ii == 0:
            plt.plot(xs, ys, 'k', alpha=.05, label='Posterior samples')
            continue
        plt.plot(xs, ys, 'k', alpha=.05)
    ys = polynomial(H.position, xs)
    plt.plot(xs, ys, 'r', linewidth=2., label='Interpolation')
    plt.plot(xs, y_true, 'b', linewidth=0.5, label='$\mathrm{f}(x) = \sin({x^2}/{10}) * x^2$')
    plt.legend()
    plt.savefig('fit.png', dpi=300)
    plt.close()

    # Print parameters
    mean = sc.mean.val
    sigma = np.sqrt(sc.var.val)
    for ii in range(len(mean)):
        print('Coefficient x**{}: {:.2E} +/- {:.2E}'.format(ii, mean[ii], sigma[ii]))


if __name__ == '__main__':
    main()



# Copyright(C) 2023 Gordian Edenhofer
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

import operator
from functools import partial, reduce
from typing import Any, List, Optional, Tuple, Union

from jax import numpy as jnp
from jax.tree_util import (
    all_leaves,
    tree_leaves,
    tree_map,
    tree_reduce,
    tree_structure,
)


class ShapeWithDtype():
    """Minimal helper class storing the shape and dtype of an object.

    Notes
    -----
    This class may not be transparent to JAX as it shall not be flattened
    itself. If used in a tree-like structure. It should only be used as leave.
    """
    def __init__(
        self, shape: Union[Tuple[()], Tuple[int], List[int], int], dtype=None
    ):
        """Instantiates a storage unit for shape and dtype.

        Parameters
        ----------
        shape : tuple or list of int
            One-dimensional sequence of integers denoting the length of the
            object along each of the object's axis.
        dtype : dtype
            Data-type of the to-be-described object.
        """
        from ..misc import is_iterable_of_non_iterables

        if isinstance(shape, int):
            shape = (shape, )
        if isinstance(shape, list):
            shape = tuple(shape)
        if not is_iterable_of_non_iterables(shape):
            ve = f"invalid shape; got {shape!r}"
            raise TypeError(ve)

        self._shape = shape
        self._dtype = jnp.float64 if dtype is None else dtype
        self._size = None

    @classmethod
    def from_leave(cls, element):
        """Convenience method for creating an instance of `ShapeWithDtype` from
        an object.

        To map a whole tree-like structure to a its shape and dtype use JAX's
        `tree_map` method like so:

            tree_map(ShapeWithDtype.from_leave, tree)

        Parameters
        ----------
        element : tree-like structure
            Object from which to take the shape and data-type.

        Returns
        -------
        swd : instance of ShapeWithDtype
            Instance storing the shape and data-type of `element`.
        """
        if not all_leaves((element, )):
            ve = "tree is not flat and still contains leaves"
            raise ValueError(ve)
        return cls(jnp.shape(element), _get_dtype(element))

    @property
    def shape(self) -> Tuple[int]:
        """Retrieves the shape."""
        return self._shape

    @property
    def dtype(self):
        """Retrieves the data-type."""
        return self._dtype

    @property
    def size(self) -> int:
        """Total number of elements."""
        if self._size is None:
            self._size = reduce(operator.mul, self.shape, 1)
        return self._size

    @property
    def ndim(self) -> int:
        return len(self.shape)

    def __len__(self) -> int:
        if self.ndim > 0:
            return self.shape[0]
        else:  # mimic numpy
            raise TypeError("len() of unsized object")

    def __eq__(self, other) -> bool:
        if not isinstance(other, ShapeWithDtype):
            return False
        else:
            return (self.shape, self.dtype) == (other.shape, other.dtype)

    def __repr__(self):
        nm = self.__class__.__name__
        return f"{nm}(shape={self.shape}, dtype={self.dtype})"

    # TODO: overlaod basic arithmetics (see `np.broadcast_shapes((1, 2), (3,
    # 1), (3, 2))`)


def _get_dtype(v: Any):
    if hasattr(v, "dtype"):
        return v.dtype
    else:
        return type(v)


def result_type(*trees):
    from numpy import result_type as result_type_np

    common_dtp = result_type_np(
        *(
            result_type_np(*(_get_dtype(v) for v in tree_leaves(tr)))
            for tr in trees
        )
    )
    return common_dtp


def _size(x):
    return x.size if hasattr(x, "size") else jnp.size(x)


def size(a, axis: Optional[int] = None) -> int:
    if axis is not None:
        raise TypeError("axis of an arbitrary tree is ill defined")
    return tree_reduce(operator.add, tree_map(_size, a))


def shape(a):
    return (size(a), )


def _like(x, dtype, shape, like, new):
    # Catch structures that plain numpy might not handle such as SWD structs
    if hasattr(x, "shape") and hasattr(x, "dtype"):
        shp = x.shape if shape is None else shape
        dtp = x.dtype if dtype is None else dtype
        return new(shape=shp, dtype=dtp)
    return like(x, dtype=dtype, shape=shape)


zeros_like = partial(
    tree_map,
    partial(_like, dtype=None, shape=None, like=jnp.zeros_like, new=jnp.zeros)
)

ones_like = partial(
    tree_map,
    partial(_like, dtype=None, shape=None, like=jnp.ones_like, new=jnp.ones)
)


def _ravel(x):
    return x.ravel() if hasattr(x, "ravel") else jnp.ravel(x)


def norm(tree, ord=2):
    """**Vector** norm.

    Notes
    -----
    This function assumes the input to be a vector, i.e. the default order `ord`
    is `2`.
    """
    from jax.numpy.linalg import norm

    def el_norm(x):
        if jnp.ndim(x) == 0:
            return jnp.abs(x)
        return norm(_ravel(x), ord=ord)

    return norm(jnp.array(tree_leaves(tree_map(el_norm, tree))), ord=ord)


def dot(a, b, *, precision=None):
    """Returns the dot product of the two vectors.

    Parameters
    ----------
    a : object
        Arbitrary, flatten-able objects.
    b : object
        Arbitrary, flatten-able objects.

    Returns
    -------
    out : float
        Dot product of vectors.
    """
    tree_of_dots = tree_map(
        lambda x, y: jnp.dot(_ravel(x), _ravel(y), precision=precision), a, b
    )
    return tree_reduce(operator.add, tree_of_dots, 0.)


matmul = dot


def vdot(a, b, *, precision=None):
    tree_of_vdots = tree_map(partial(jnp.vdot, precision=precision), a, b)
    return tree_reduce(jnp.add, tree_of_vdots, 0.)


def _conj(x):
    return x.conj() if hasattr(x, "conj") else jnp.conj(x)


def conjugate(a):
    """Returns the complex conjugate, component- and element-wise.

    Parameters
    ----------
    a : object
        Arbitrary, flatten-able objects.

    Returns
    -------
    out : object
        The complex conjugate of `a`, with same shape and dtype as `a`.
    """
    return tree_map(_conj, a)


conj = conjugate


def where(condition, x, y):
    """Selects a pytree based on the condition which can be a pytree itself.

    Notes
    -----
    If `condition` is not a pytree, then a partially evaluated selection is
    simply mapped over `x` and `y` without actually broadcasting `condition`.
    """
    from itertools import repeat

    import numpy as np

    ts_c = tree_structure(condition)
    ts_x = tree_structure(x)
    ts_y = tree_structure(y)
    ts_max = (ts_c, ts_x, ts_y)[np.argmax(
        [ts_c.num_nodes, ts_x.num_nodes, ts_y.num_nodes]
    )]

    if ts_x.num_nodes < ts_max.num_nodes:
        if ts_x.num_nodes > 1:
            raise ValueError("can not broadcast LHS")
        x = ts_max.unflatten(repeat(x, ts_max.num_leaves))
    if ts_y.num_nodes < ts_max.num_nodes:
        if ts_y.num_nodes > 1:
            raise ValueError("can not broadcast RHS")
        y = ts_max.unflatten(repeat(y, ts_max.num_leaves))

    if ts_c.num_nodes < ts_max.num_nodes:
        if ts_c.num_nodes > 1:
            raise ValueError("can not map condition")
        return tree_map(partial(jnp.where, condition), x, y)
    return tree_map(jnp.where, condition, x, y)


def _unary_reduction(jax_f, name=None):
    name = name if name is not None else jax_f.__name__

    def _safe_uni(a):
        return jax_f(jnp.atleast_1d(a))

    def _red(a, b):
        return jax_f(jnp.array([a, b]))

    def unary_reduction(a):
        return tree_reduce(_red, tree_map(_safe_uni, a))

    unary_reduction.__name__ = name
    return unary_reduction


sum = _unary_reduction(jnp.sum)
min = _unary_reduction(jnp.min)
max = _unary_reduction(jnp.max)
any = _unary_reduction(jnp.any)
all = _unary_reduction(jnp.all)



#!/usr/bin/env python3

import pytest

pytest.importorskip("jax")

from functools import partial

import jax
import jax.numpy as jnp
from jax import random
from jax.tree_util import tree_map
from numpy.testing import assert_allclose

import nifty8.re as jft

pmp = pytest.mark.parametrize

jax.config.update("jax_enable_x64", True)


def f(u, v):
    return jnp.exp(u @ v)


@pmp("map", (jft.smap, jft.lmap))
@pmp("in_axes", ((0, None), (1, None), (None, 0), (0, 0), (1, 1)))
@pmp("out_axes", (0, ))
@pmp("seed", (32, 42, 43))
def test_map_f(map, in_axes, out_axes, seed):
    fixed_shp = 3
    batched_shp = 4
    key = random.PRNGKey(seed)

    vf = jax.vmap(f, in_axes=in_axes, out_axes=out_axes)
    sf = map(f, in_axes=in_axes, out_axes=out_axes)
    inp = []
    for i, k in zip(in_axes, random.split(key, len(in_axes))):
        if i is None:
            shp = (fixed_shp, )
        elif i == 0:
            shp = (batched_shp, fixed_shp)
        else:
            assert i == 1
            shp = (fixed_shp, batched_shp)
        inp.append(random.normal(k, shape=shp))
    assert_allclose(vf(*inp), sf(*inp), atol=1e-14, rtol=1e-14)


def g(u, v):
    return u, jnp.exp(u @ v)


@pmp("map", (jft.smap, jft.lmap))
@pmp("in_axes,out_axes", (((0, None), (0, 0)), ((None, 1), (None, 0))))
@pmp("seed", (32, 42, 43))
def test_map_g(map, in_axes, out_axes, seed):
    fixed_shp = 3
    batched_shp = 4
    key = random.PRNGKey(seed)

    vf = jax.vmap(g, in_axes=in_axes, out_axes=out_axes)
    sf = map(g, in_axes=in_axes, out_axes=out_axes)
    inp = []
    for i, k in zip(in_axes, random.split(key, len(in_axes))):
        if i is None:
            shp = (fixed_shp, )
        elif i == 0:
            shp = (batched_shp, fixed_shp)
        else:
            assert i == 1
            shp = (fixed_shp, batched_shp)
        inp.append(random.normal(k, shape=shp))
    v = vf(*inp)
    s = sf(*inp)
    tree_map(partial(assert_allclose, atol=1e-14, rtol=1e-14), v, s)


def h(u, v, w):
    return v + w.sum() - u.sum(), w * jnp.sin(u.sum() * v.sum()
                                             ), u + v.sum() * w.sum()


@pmp("map", (jft.smap, jft.lmap))
@pmp("in_axes", (0, 1, 2))
@pmp("out_axes", (0, 1, 2))
@pmp("seed", (32, 42, 43))
def test_map_h(map, in_axes, out_axes, seed):
    key = random.PRNGKey(seed)

    vf = jax.vmap(h, in_axes=in_axes, out_axes=out_axes)
    sf = map(h, in_axes=in_axes, out_axes=out_axes)
    inp = [random.normal(k, shape=(1, 2, 3)) for k in random.split(key, 3)]
    v = vf(*inp)
    s = sf(*inp)
    tree_map(partial(assert_allclose, atol=1e-14, rtol=1e-14), v, s)



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

import pytest

pytest.importorskip("jax")

from functools import partial

import jax
import jax.numpy as jnp
import numpy as np
from jax import random
from numpy.testing import assert_allclose, assert_array_equal

import nifty8.re as jft

jax.config.update("jax_enable_x64", True)
pmp = pytest.mark.parametrize


def wiener_filter_posterior(data, response, noise_covariance, prior_covariance):
    r"""Compute the posterior mean for the
    [Wiener filter](https://en.wikipedia.org/wiki/Information_field_theory#Generalized_Wiener_filter)
    setting.

    The maximally plausible $\rho_m$ is
    \begin{equation}
        \rho_m = (R^T N^{-1}R + S^{-1})^{-1} R^T N^{-1} \mathcal{D}_A
    \end{equation}
    with $N$ a matrix with $\sigma_{A,i}$ on its diagonal, $\blacksquare^T$ the
    [matrix-transpose](https://en.wikipedia.org/wiki/Transpose), and
    $\blacksquare^{-1}$ the
    [matrix-inverse](https://en.wikipedia.org/wiki/Invertible_matrix).
    """
    noise_covariance = np.diag(
        noise_covariance
    ) if noise_covariance.ndim == 1 else noise_covariance
    noise_cov_inv = np.linalg.inv(noise_covariance)
    post_cov_inv = response.T @ noise_cov_inv @ response + np.linalg.inv(
        prior_covariance
    )
    post_cov = np.linalg.inv(post_cov_inv)
    return post_cov @ response.T @ noise_cov_inv @ data, post_cov


def los_response(end, shape, extent=(0, 1.), start=0.):
    """Compute the line-of-sight response matrix for a given shape and extent.
    """
    if len(shape) > 1:
        raise ValueError("only 1D shapes are supported.")
    start, end = np.broadcast_arrays(start, end)
    if np.any(start > end):
        raise ValueError("start must be smaller than end.")

    # Convert to dimensionless pixel indices
    start = (start - extent[0]) / (extent[1] - extent[0]) * (shape[0] - 1)
    end = (end - extent[0]) / (extent[1] - extent[0]) * (shape[0] - 1)
    response = []
    for i in range(shape[0]):
        # print(i, start, (i + 0.5 - start).clip(0., 1.))
        # Assume the zeroth pixel is centered at 0.0
        l = np.minimum(
            (i + 0.5 - start).clip(0., 1.), (end - i + 0.5).clip(0., 1.)
        )
        response.append(l)
    return np.stack(response, axis=1)


def signal_cov(shape, extent=(0, 1.), scale=1., lengthscale=0.1):
    from scipy.spatial import distance_matrix

    if len(shape) > 1:
        raise ValueError("only 1D shapes are supported.")
    p = np.linspace(*extent, num=shape[0], endpoint=True)
    d = distance_matrix(p.reshape(-1, 1), p.reshape(-1, 1))
    return scale * (np.exp(-(d / lengthscale)**2) + 1e-6 * np.eye(shape[0]))


@pmp("seed", (12, 42))
def test_mgvi_wiener_filter_consistency(
    seed,
    shape=(16, ),
    extent=(0., 1.),
    n_los=14,
    n_mgvi_cov_samples=16_000,
    n_vi_samples=16,
):
    key = random.PRNGKey(seed)
    size = np.prod(shape)

    key, k_r, k_p, k_n = random.split(key, 4)
    prior_cov = signal_cov(shape, extent=extent)
    prior_lsm = np.linalg.cholesky(prior_cov)
    dust_synth = prior_lsm @ random.normal(k_p, shape=shape)
    dust_synth = np.exp(dust_synth)  # Fake positivity; THIS VOILATES THE PRIOR
    endpoints = random.uniform(
        k_r, minval=extent[0], maxval=extent[1], shape=(n_los, )
    )
    response = los_response(end=endpoints, shape=shape, extent=extent)
    k_n1, k_n2 = random.split(k_n)
    noise_cov = np.mean(dust_synth) * (
        1 + random.uniform(k_n1, minval=0.1, maxval=0.5, shape=(n_los, ))
    )
    data = response @ dust_synth
    data += np.sqrt(noise_cov) * random.normal(k_n2, shape=(n_los, ))
    # plt.scatter(endpoints, data); plt.show()

    # Retrieve Wiener Filter posterior (even though model draws data slightly
    # differently)
    post_mean, post_cov = wiener_filter_posterior(
        data, response, noise_cov, prior_cov
    )
    # plt.plot(np.array((dust_synth, post_mean)).T); plt.show()
    # plt.matshow(post_cov); plt.colorbar(); plt.show()

    # Retrieve the equivalent Wiener filter solution for a standardized prior
    post_mean_st, post_cov_st = wiener_filter_posterior(
        data, response @ prior_lsm, noise_cov, np.eye(size)
    )
    post_mean_alt = prior_lsm @ post_mean_st
    post_cov_alt = prior_lsm @ post_cov_st @ prior_lsm.T
    # Check for consistency with the "normal" Wiener filter solution
    assert_allclose(post_mean_alt, post_mean, atol=1e-8, rtol=1e-8)
    assert_allclose(post_cov_alt, post_cov, atol=1e-8, rtol=1e-8)

    # Retrieve posterior via NIFTy
    lh = jft.Gaussian(
        data, noise_cov_inv=partial(jnp.multiply, np.reciprocal(noise_cov))
    )
    forward = partial(jnp.matmul, prior_lsm)
    lh = lh.amend(partial(jnp.matmul, response)).amend(forward)

    draw_linear_kwargs = dict(cg_kwargs=dict(absdelta=1e-10, maxiter=size))
    draw_linear_residual = partial(
        jft.draw_linear_residual, lh, **draw_linear_kwargs
    )
    draw_linear_residual = jft.smap(draw_linear_residual, in_axes=(None, 0))
    # Compare metric of Likelihood + Prior against true posterior covariance
    key, sk = random.split(key)
    some_positions = (
        random.normal(sk, shape=shape), jnp.ones(shape), jnp.zeros(shape)
    )
    key, k_smpl = random.split(key)
    for pos in some_positions:
        probe = jnp.zeros(shape)
        lh_metric = jax.vmap(
            lambda i: lh.metric(pos, probe.at[i].set(1.)), out_axes=1
        )(jnp.arange(size))
        lh_metric_diy = prior_lsm.T @ response.T @ np.diag(
            1. / noise_cov
        ) @ response @ prior_lsm
        assert_allclose(lh_metric, lh_metric_diy, atol=1e-10, rtol=1e-12)
        mgvi_post_cov = jnp.linalg.inv(lh_metric + jnp.eye(size))
        # Compare posterior cov in space of standardized prior
        assert_allclose(mgvi_post_cov, post_cov_st, atol=1e-12, rtol=1e-14)
        # plt.matshow(mgvi_post_cov); plt.colorbar(); plt.show()

        # Test sampling approach
        residual_draw, info = draw_linear_residual(
            pos, random.split(k_smpl, n_mgvi_cov_samples)
        )
        assert_array_equal(info, 0)
        residual_draw = jax.vmap(forward)(residual_draw)
        tol = (np.diag(post_cov) / n_mgvi_cov_samples).sum()
        assert_allclose(residual_draw.mean(axis=0)**2, 0., atol=tol)
        smpl_post_cov = np.cov(residual_draw, rowvar=False)
        assert_allclose(
            np.diag(smpl_post_cov), np.diag(post_cov), atol=tol**0.5
        )
        assert_allclose(smpl_post_cov, post_cov, atol=tol**0.5)

    kl_kwargs = dict(minimize_kwargs=dict(name="M", xtol=1e-10, maxiter=35))

    # MGVI test
    key, sk = random.split(key)
    samples_opt, _ = jft.optimize_kl(
        lh,
        pos,
        key=sk,
        n_total_iterations=1,
        n_samples=n_vi_samples,
        draw_linear_kwargs=draw_linear_kwargs,
        kl_kwargs=kl_kwargs,
        sample_mode="linear_sample",
        odir=None,
    )
    approx_post_mean = jax.vmap(forward)(samples_opt.samples).mean(axis=0)
    assert_allclose(post_mean, approx_post_mean, atol=1e-9, rtol=1e-10)

    delta = 1e-3
    nonlinearly_update_kwargs = dict(
        minimize_kwargs=dict(
            name="SN",
            xtol=delta,
            cg_kwargs=dict(name=None, miniter=2),
            maxiter=5,
        )
    )

    # geoVI test
    key, sk = random.split(key)
    samples_opt, _ = jft.optimize_kl(
        lh,
        pos,
        key=sk,
        n_total_iterations=1,
        n_samples=n_vi_samples,
        draw_linear_kwargs=draw_linear_kwargs,
        nonlinearly_update_kwargs=nonlinearly_update_kwargs,
        kl_kwargs=kl_kwargs,
        sample_mode="nonlinear_sample",
        odir=None,
    )
    approx_post_mean = jax.vmap(forward)(samples_opt.samples).mean(axis=0)
    assert_allclose(post_mean, approx_post_mean, atol=1e-10, rtol=1e-10)


if __name__ == "__main__":
    test_mgvi_wiener_filter_consistency(
        42,
        shape=(16, ),
        extent=(0., 1.),
        n_los=14,
        n_mgvi_cov_samples=16_000,
        n_vi_samples=16,
    )



#!/usr/bin/env python3

# Copyright(C) 2013-2021 Max-Planck-Society
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

#%%
from jax import numpy as jnp
from jax import lax, random
import jax
from jax.config import config
import matplotlib
import matplotlib.pyplot as plt

import nifty8.re as jft

config.update("jax_enable_x64", True)

matplotlib.rcParams['figure.figsize'] = (10, 7)

#%%
dims = (512, )
#datadims = (4,)
loglogslope = 2.
power_spectrum = lambda k: 1. / (k**loglogslope + 1.)
modes = jnp.arange((dims[0] / 2) + 1., dtype=float)
harmonic_power = power_spectrum(modes)
harmonic_power = jnp.concatenate((harmonic_power, harmonic_power[-2:0:-1]))

#%%
correlated_field = lambda x: jft.correlated_field.hartley(
    # x is a signal in fourier space
    # each modes amplitude gets multiplied by it's harmonic_power
    # and the whole signal is transformed back
    harmonic_power * x
)

# %% [markdown]
# signal_response = lambda x: jnp.exp(1. + correlated_field(x))
signal_response = lambda x: correlated_field(x)
# The signal response is $ \vec{d} = \begin{pmatrix} 1 \\ 1 \\ 1 \\ 1 \end{pmatrix} \cdot s + \vec{n} $ where $s \in \mathbb{R}$ and $\vec{n} \sim \mathcal{G}(0, N)$
# signal_response = lambda x: jnp.ones(shape=datadims) * x
# ???
noise_cov_inv_sqrt = lambda x: 1.**-1 * x

#%%
# create synthetic data
seed = 43
key = random.PRNGKey(seed)
key, subkey = random.split(key)
# normal random fourier amplitude
pos_truth = random.normal(shape=dims, key=subkey)
signal_response_truth = signal_response(pos_truth)
key, subkey = random.split(key)
# 1. / noise_cov_inv_sqrt(jnp.ones(dims)) becomes the standard deviation of the noise gaussian
noise_truth = 1. / noise_cov_inv_sqrt(jnp.ones(dims)
                                     ) * random.normal(shape=dims, key=subkey)
data = signal_response_truth + noise_truth

#%%
plt.plot(signal_response_truth, label='signal response')
#plt.plot(noise_truth, label='noise', linewidth=0.5)
plt.plot(data, 'k.', label='noisy data', markersize=4.)
plt.xlabel('real space domain')
plt.ylabel('field value')
plt.legend()
plt.title("signal and data")
plt.show()


#%%
def Gaussian(data, noise_cov_inv_sqrt):
    # Simple but not very generic Gaussian energy
    # primals
    def hamiltonian(primals):
        p_res = primals - data
        # TODO: is this the weighting with noies amplitude thing again?
        l_res = noise_cov_inv_sqrt(p_res)
        return 0.5 * jnp.sum(l_res**2)

    return jft.Likelihood(hamiltonian, )


# negative log likelihood
nll = Gaussian(data, noise_cov_inv_sqrt) @ signal_response

#%%
ham = jft.StandardHamiltonian(likelihood=nll)
ham_gradient = jax.grad(ham)


# %% [markdown]
def plot_mean_and_stddev(ax, samples, mean_of_r=None, truth=False, **kwargs):
    signal_response_of_samples = lax.map(signal_response, samples)
    if mean_of_r == None:
        mean_of_signal_response = jnp.mean(signal_response_of_samples, axis=0)
    else:
        mean_of_signal_response = mean_of_r
    mean_label = kwargs.pop('mean_label', 'sample mean of signal response')
    ax.plot(mean_of_signal_response, label=mean_label)
    std_dev_of_signal_response = jnp.std(signal_response_of_samples, axis=0)
    if truth:
        ax.plot(signal_response_truth, label="truth")
    ax.fill_between(
        jnp.arange(len(mean_of_signal_response)),
        y1=mean_of_signal_response - std_dev_of_signal_response,
        y2=mean_of_signal_response + std_dev_of_signal_response,
        color='grey',
        alpha=0.5
    )
    title = kwargs.pop('title', 'position samples')
    if title is not None:
        ax.set_title(title)
    xlabel = kwargs.pop('xlabel', 'position')
    if xlabel is not None:
        ax.set_xlabel(xlabel)
    ylabel = kwargs.pop('ylabel', 'signal response')
    if ylabel is not None:
        ax.set_ylabel(ylabel)
    ax.legend(loc='lower right', fontsize=8)


#%%
key, subkey = random.split(key)
initial_position = random.uniform(key=subkey, shape=pos_truth.shape)

sampler = jft.HMCChain(
    potential_energy=ham,
    inverse_mass_matrix=1.,
    position_proto=initial_position,
    step_size=0.05,
    num_steps=128,
)

chain, _ = sampler.generate_n_samples(
    42, initial_position, num_samples=30, save_intermediates=True
)
print(f"acceptance ratio: {chain.acceptance}")

# %%
plot_mean_and_stddev(plt.gca(), chain.samples, truth=True)
plt.title("HMC position samples")
plt.show()

# %% [markdown]
# # NUTS
jft.hmc._DEBUG_STORE = []

sampler = jft.NUTSChain(
    position_proto=initial_position,
    potential_energy=ham,
    inverse_mass_matrix=1.,
    # 0.9193 # integrates to ~3-7, very smooth sample mean
    # 0.8193 # integrates to depth ~22, very noisy sample mean
    step_size=0.05,
    max_tree_depth=17,
)

chain, _ = sampler.generate_n_samples(
    42, initial_position, num_samples=30, save_intermediates=True
)
plt.hist(chain.depths, bins=jnp.arange(sampler.max_tree_depth + 2))
plt.title('NUTS tree depth histogram')
plt.xlabel('tree depth')
plt.ylabel('count')
plt.show()

# %%
plot_mean_and_stddev(plt.gca(), chain.samples, truth=True)
plt.title("NUTS position samples")
plt.show()

# %%
if jft.hmc._DEBUG_FLAG:
    debug_pos = jnp.array(jft.hmc._DEBUG_STORE)[:, 0, :]

    for idx, dbgp in enumerate(debug_pos):
        plt.plot(signal_response(dbgp), label=f'{idx}', alpha=0.1)
    #plt.legend()

    # %%
    debug_pos_x = jnp.array(jft.hmc._DEBUG_STORE)[:, 0, 0]
    debug_pos_y = jnp.array(jft.hmc._DEBUG_STORE)[:, 0, 1]
    for idx, dbgp in enumerate(debug_pos):
        plt.scatter(debug_pos_x, debug_pos_y, s=0.1, color='k')
    #plt.legend()
    plt.show()

# %%[markdown]
# # 1D position and momentum time series
if chain.samples[0].shape == (1, ):
    plt.plot(chain.samples, label='position')
    #plt.plot(momentum_samples, label='momentum', linewidth=0.2)
    #plt.plot(unintegrated_momenta, label='unintegrated momentum', linewidth=0.2)
    plt.title('position and momentum time series')
    plt.xlabel('time')
    plt.ylabel('position, momentum')
    plt.legend()
    plt.show()

# %% [markdown]
# # energy time series
potential_energies = lax.map(ham, chain.samples)
kinetic_energies = jnp.sum(chain.trees.proposal_candidate.momentum**2, axis=1)
#rejected_potential_energies = lax.map(ham, rejected_position_samples)
#rejected_kinetic_energies = jnp.sum(rejected_momentum_samples**2, axis=1)
plt.plot(potential_energies, label='pot')
plt.plot(kinetic_energies, label='kin', linewidth=1)
plt.plot(kinetic_energies + potential_energies, label='total', linewidth=1)
#plt.plot(rejected_potential_energies , label='rejected_pot')
#plt.plot(rejected_kinetic_energies , label='rejected_kin', linewidth=2)
#plt.plot(rejected_kinetic_energies + rejected_potential_energies, label='rejected_total', linewidth=0.2)
plt.title('NUTS energy time series')
plt.xlabel('time')
plt.ylabel('energy')
plt.yscale('log')
plt.legend()
plt.show()

# %% [markdown]
# # Wiener Filter

# jax.linear_transpose for R^\dagger
# square noise_sqrt_inv ... for N^-1
# S is unit due to StandardHamiltonian
# jax.scipy.sparse.linalg.cg for D

# signal_response(s) is only needed for shape of data space
_impl_signal_response_dagger = jax.linear_transpose(signal_response, pos_truth)
signal_response_dagger = lambda d: _impl_signal_response_dagger(d)[0]
# noise_cov_inv_sqrt is diagonal
noise_cov_inv = lambda d: noise_cov_inv_sqrt(noise_cov_inv_sqrt(d))

# signal prior covariance S is assumed to be unit (see StandardHamiltonian)
# the tranposed function wierdly returns a (1,)-tuple which we unpack right here
D_inv = lambda s: s + signal_response_dagger(noise_cov_inv(signal_response(s)))

j = signal_response_dagger(noise_cov_inv(data))

m, _ = jax.scipy.sparse.linalg.cg(D_inv, j)

# %%

# TODO fix labels
plt.plot(signal_response(m), label='signal response of mean')
plt.plot(signal_response_truth, label='true signal response')
plt.legend()
plt.title('Wiener Filter')
plt.show()


# %%
def sample_from_d_inv(key):
    s_inv_key, rnr_key = random.split(key)
    s_inv_smpl = random.normal(s_inv_key, pos_truth.shape)
    # random.normal sample from dataspace and then R^\dagger \sqrt{N^{-1}}
    # jax.eval_shape(signal_response, pos_truth)
    rnr_smpl = signal_response_dagger(
        noise_cov_inv_sqrt(random.normal(rnr_key, signal_response_truth.shape))
    )
    return s_inv_smpl + rnr_smpl


def sample_from_d(key):
    d_inv_smpl = sample_from_d_inv(key)
    # TODO: what to do here?
    smpl, _ = jft.cg(D_inv, d_inv_smpl, maxiter=32)
    return smpl


wiener_samples = jnp.array(
    list(map(lambda key: sample_from_d(key) + m, random.split(key, 30)))
)

# %%
subplots = (3, 1)
fig_height_pt = 541  # pt
#fig_width_pt = 360 # pt
inches_per_pt = 1 / 72.27
fig_height_in = 1. * fig_height_pt * inches_per_pt
fig_width_in = fig_height_in / 0.618 * (subplots[1] / subplots[0])
fig_dims = (fig_width_in, fig_height_in)

fig, (ax_raw, ax_nuts, ax_wiener) = plt.subplots(
    subplots[0], subplots[1], sharex=True, sharey=False, figsize=fig_dims
)

ax_raw.plot(signal_response_truth, label='true signal response')
ax_raw.plot(data, 'k.', label='noisy data', markersize=2.)
#ax_raw.set_xlabel('position')
ax_raw.set_ylabel('signal response')
ax_raw.set_title("signal and data")
ax_raw.legend(fontsize=8)

plot_mean_and_stddev(
    ax_nuts,
    chain.samples,
    truth=True,
    title="NUTS",
    xlabel=None,
    mean_label='sample mean'
)
plot_mean_and_stddev(
    ax_wiener,
    wiener_samples,
    mean_of_r=signal_response(m),
    truth=True,
    title="Wiener Filter",
    mean_label='exact posterior mean'
)

fig.tight_layout()

plt.savefig('wiener.pdf', bbox_inches='tight')
print("final plot saved as wiener.pdf")



# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

import operator
from copy import deepcopy
from pprint import pformat

from jax import numpy as jnp
from jax.tree_util import (
    register_pytree_node_class, tree_leaves, tree_map, tree_structure
)

from .vector_math import matmul, max, min, size, sum


def _copy(obj):
    return obj.copy() if hasattr(obj, "copy") else deepcopy(obj)


def _value_op(op, name=None):
    def value_call(lhs, *args, **kwargs):
        return op(lhs.tree, *args, **kwargs)

    name = op.__name__ if name is None else name
    value_call.__name__ = f"__{name}__"
    return value_call


def _unary_op(op, name=None):
    def unary_call(lhs):
        return tree_map(op, lhs)

    name = op.__name__ if name is None else name
    unary_call.__name__ = f"__{name}__"
    return unary_call


def _broadcast_binary_op(op, lhs, rhs):
    from itertools import repeat

    ts_lhs = tree_structure(lhs)
    ts_rhs = tree_structure(rhs)
    # Catch non-objects scalars and 0d array-likes with a `ndim` property
    if jnp.isscalar(lhs) or getattr(lhs, "ndim", -1) == 0:
        lhs = ts_rhs.unflatten(repeat(lhs, ts_rhs.num_leaves))
    elif jnp.isscalar(rhs) or getattr(rhs, "ndim", -1) == 0:
        rhs = ts_lhs.unflatten(repeat(rhs, ts_lhs.num_leaves))
    elif ts_lhs.num_nodes != ts_rhs.num_nodes:
        ve = f"invalid binary operation {op} for {ts_lhs!r} and {ts_rhs!r}"
        raise ValueError(ve)
    return tree_map(op, lhs, rhs)


def _binary_op(op, name=None):
    def binary_call(lhs, rhs):
        return _broadcast_binary_op(op, lhs, rhs)

    name = op.__name__ if name is None else name
    binary_call.__name__ = f"__{name}__"
    return binary_call


def _rev_binary_op(op, name=None):
    def binary_call(lhs, rhs):
        return _broadcast_binary_op(op, rhs, lhs)

    name = op.__name__ if name is None else name
    binary_call.__name__ = f"__r{name}__"
    return binary_call


def _fwd_rev_binary_op(op, name=None):
    return (_binary_op(op, name=name), _rev_binary_op(op, name=name))


@register_pytree_node_class
class Vector():
    """Value storage for arbitrary objects with added numerics."""
    def __init__(self, tree):
        """Instantiates a vector.

        Parameters
        ----------
        tree : object
            Arbitrary, flatten-able objects.
        """
        self._tree = tree

    def tree_flatten(self):
        return ((self._tree, ), None)

    @classmethod
    def tree_unflatten(cls, _, children):
        return cls(*children)

    @property
    def tree(self):
        """Retrieves a **view** of the vector's values."""
        return self._tree

    def __len__(self):
        return size(self)

    @property
    def size(self):
        return len(self)

    @property
    def shape(self):
        return (len(self), )

    def copy(self):
        return tree_map(_copy, self)

    def ravel(self):
        return self

    def __repr__(self):
        rep = pformat(self.tree).replace("\n", "\n\t").strip()
        s = f"{self.__class__.__name__}(\n\t{rep}\n)"
        s = s.replace("\n", "").replace("\t", "") if s.count("\n") <= 2 else s
        return s

    def __str__(self):
        return repr(self)

    __bool__ = _value_op(bool)

    def __hash__(self):
        return hash(tuple(tree_leaves(self)))

    # NOTE, this partly redundant code could be abstracted away using
    # `setattr`. However, static code analyzers will not be able to infer the
    # properties then.

    __add__, __radd__ = _fwd_rev_binary_op(operator.add)
    __sub__, __rsub__ = _fwd_rev_binary_op(operator.sub)
    __mul__, __rmul__ = _fwd_rev_binary_op(operator.mul)
    __truediv__, __rtruediv__ = _fwd_rev_binary_op(operator.truediv)
    __floordiv__, __rfloordiv__ = _fwd_rev_binary_op(operator.floordiv)
    __pow__, __rpow__ = _fwd_rev_binary_op(operator.pow)
    __mod__, __rmod__ = _fwd_rev_binary_op(operator.mod)
    __matmul__ = __rmatmul__ = matmul  # arguments of matmul commute

    def __divmod__(self, other):
        return self // other, self % other

    def __rdivmod__(self, other):
        return other // self, other % self

    __or__, __ror__ = _fwd_rev_binary_op(operator.or_, "or")
    __xor__, __rxor__ = _fwd_rev_binary_op(operator.xor)
    __and__, __rand__ = _fwd_rev_binary_op(operator.and_, "and")
    __lshift__, __rlshift__ = _fwd_rev_binary_op(operator.lshift)
    __rshift__, __rrshift__ = _fwd_rev_binary_op(operator.rshift)

    __lt__ = _binary_op(operator.lt)
    __le__ = _binary_op(operator.le)
    __eq__ = _binary_op(operator.eq)
    __ne__ = _binary_op(operator.ne)
    __ge__ = _binary_op(operator.ge)
    __gt__ = _binary_op(operator.gt)

    __neg__ = _unary_op(operator.neg)
    __pos__ = _unary_op(operator.pos)
    __abs__ = _unary_op(operator.abs)
    __invert__ = _unary_op(operator.invert)

    conj = conjugate = _unary_op(jnp.conj)
    real = property(_unary_op(jnp.real))
    imag = property(_unary_op(jnp.imag))
    dot = matmul

    def max(self):
        return max(self)

    def min(self):
        return min(self)

    def sum(self):
        return sum(self)

    __getitem__ = _value_op(operator.getitem)
    __contains__ = _value_op(operator.contains)
    __iter__ = _value_op(iter)



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

import jax
import numpy as np
import pytest
import jax.numpy as jnp
from jax import random, vmap
from numpy.testing import assert_allclose

import nifty8.re as jft

pmp = pytest.mark.parametrize


def _call(process, seed):
    p = process.init(random.PRNGKey(seed))
    process = jax.jit(process)
    assert np.all(np.isfinite(process(p)))


@pmp('x0', [0.1, -2.2])
@pmp('dts', [np.ones((10, )) * 0.2, np.array([0.2, 0.5, 0.1])])
@pmp('gamma', [0.1, 0.5, 0.2])
def test_decay_ornstein_uhlenbeck(x0, dts, gamma):
    ts = np.zeros(dts.size + 1)
    ts[1:] = np.cumsum(dts)
    res = np.exp(-gamma * ts) * x0
    gp = jft.OrnsteinUhlenbeckProcess(1., gamma, dts, x0=x0)
    myres = gp(jft.zeros_like(gp.domain))
    assert_allclose(res, myres)


@pmp('x0', [0.1, -1.2])
@pmp('dts', [np.ones((10, )) * 0.2, np.array([0.2, 0.5, 0.1])])
@pmp('seed', [
    42,
])
def test_const_wiener(x0, dts, seed):
    gp = jft.WienerProcess(x0, 1., dts)
    myres = gp(jft.zeros_like(gp.domain))
    assert_allclose(myres, x0)
    # Compare to cumsum
    p = gp.init(random.PRNGKey(seed))
    res = x0 + np.cumsum(p['wp'] * np.sqrt(dts))
    gpres = gp(p)
    assert_allclose(res, gpres[1:])

    amp = np.sqrt(dts)
    res2 = jft.gauss_markov.scalar_gauss_markov_process(p['wp'], x0, 1., amp)
    assert_allclose(res2, gpres)


@pmp('x0', [np.array([0.1, 0.5]), np.array([-1.2, 0.7])])
@pmp('dts', [np.ones((10, )) * 0.2, np.array([0.2, 0.5, 0.1])])
def test_drift_integrated_wiener(x0, dts):
    ts = np.zeros(dts.size + 1)
    ts[1:] = np.cumsum(dts)
    res = x0[1] * ts + x0[0]
    gp = jft.IntegratedWienerProcess(x0, 1., dts)
    myres = gp(jft.zeros_like(gp.domain))
    # Check if it is a straight line
    assert_allclose(myres[:, 0], res)
    # Check that derivative is const.
    assert_allclose(myres[:, 1], x0[1])


@pmp('dts', [np.ones((10, )) * 0.2, np.array([0.2, 0.5, 0.1])])
def test_iwp_cumsum_vs_fori(dts):
    """Implements the (generalized) Integrated Wiener process (IWP)."""
    x0 = np.array([0.1, 0.5])
    sigma = 1.
    asperity = 0.1
    gp = jft.IntegratedWienerProcess(x0, sigma, dts, asperity=asperity)

    rnd = gp.init(random.PRNGKey(10))
    res = gp(rnd)
    xi = rnd['iwp']

    def drift_amp(d, sig, asp):
        drift = jnp.array([[1., d], [0., 1.]])
        amp = jnp.array([[jnp.sqrt(d**2 / 12. + asp), d / 2.], [0., 1.]])
        amp *= sig * jnp.sqrt(d)
        return drift, amp

    axs = (0, None, None)
    drift, amp = vmap(drift_amp, axs, (0, 0))(dts, sigma, asperity)
    res2 = jft.gauss_markov.discrete_gauss_markov_process(xi, x0, drift, amp)

    assert_allclose(res, res2)


@pmp('x0', [1., (0., 1.5), jft.UniformPrior(-1., 1., name='x0')])
@pmp('sigma', [0.3, (1., 0.8), jft.UniformPrior(0.1, 1.1, name='sigma')])
@pmp('name', [
    'wiener',
])
@pmp('dt, N_steps', [(0.1, 10), (np.arange(10), None), (np.ones(5), 5)])
@pmp('seed', [3, 4])
def test_wiener_eval(x0, sigma, dt, name, N_steps, seed):
    gp = jft.WienerProcess(x0, sigma, dt, name=name, N_steps=N_steps)
    _call(gp, seed)


@pmp(
    'x0', [
        np.array([0.3, -0.4]), (0., 1.5),
        jft.UniformPrior(-1., 1., name='x0', shape=(2, ))
    ]
)
@pmp('sigma', [0.3, (1., 0.8), jft.UniformPrior(0.1, 1.1, name='sigma')])
@pmp('name', [
    'intwiener',
])
@pmp('dt, N_steps', [(0.1, 10), (np.arange(10), None), (np.ones(5), 5)])
@pmp(
    'asperity',
    [None, 0.2,
     (1.1, 0.5), jft.UniformPrior(0.1, 2.2, name='asp')]
)
@pmp('seed', [3, 4])
def test_integrated_wiener_eval(x0, sigma, dt, name, N_steps, asperity, seed):
    gp = jft.IntegratedWienerProcess(
        x0, sigma, dt, name=name, N_steps=N_steps, asperity=asperity
    )
    _call(gp, seed)


@pmp('x0', [None, 1., (0., 1.5), jft.UniformPrior(-1., 1., name='x0')])
@pmp('sigma', [0.3, (1., 0.8), jft.UniformPrior(0.1, 1.1, name='sigma')])
@pmp('name', [
    'ornsteinup',
])
@pmp('dt, N_steps', [(0.1, 10), (np.arange(10), None), (np.ones(5), 5)])
@pmp('gamma', [0.2, (1.1, 0.5), jft.UniformPrior(0.1, 2.2, name='asp')])
@pmp('seed', [3, 4])
def test_ornstein_uhlenbeck_eval(x0, sigma, dt, name, N_steps, gamma, seed):
    gp = jft.OrnsteinUhlenbeckProcess(
        sigma, gamma, dt, name=name, x0=x0, N_steps=N_steps
    )
    _call(gp, seed)



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

import pytest

pytest.importorskip("jax")

from functools import partial, reduce

import jax
import jax.numpy as jnp
import numpy as np
from jax import random
from jax.tree_util import tree_map
from numpy.testing import assert_allclose, assert_array_equal

import nifty8.re as jft
from nifty8.re.optimize_kl import concatenate_zip

pmp = pytest.mark.parametrize


def random_draw(key, shape, dtype, method):
    def _isleaf(x):
        if isinstance(x, tuple):
            return bool(
                reduce(lambda a, b: a * b, (isinstance(ii, int) for ii in x))
            )
        return False

    swd = tree_map(
        lambda x: jft.ShapeWithDtype(x, dtype), shape, is_leaf=_isleaf
    )
    return jft.random_like(key, jft.Vector(swd), method)


def random_noise_std_inv(key, shape):
    diag = 1. / random_draw(key, shape, float, random.exponential)

    def noise_std_inv(tangents):
        return diag * tangents

    return noise_std_inv


LH_INIT = (
    (
        jft.Gaussian,
        {
            "data": partial(random_draw, dtype=float, method=random.normal),
            "noise_std_inv": random_noise_std_inv
        },
        partial(random_draw, dtype=float, method=random.normal),
    ), (
        jft.StudentT,
        {
            "data": partial(random_draw, dtype=float, method=random.normal),
            "dof": partial(random_draw, dtype=float, method=random.exponential),
            "noise_std_inv": random_noise_std_inv
        },
        partial(random_draw, dtype=float, method=random.normal),
    ), (
        jft.Poissonian,
        {
            "data":
                partial(
                    random_draw,
                    dtype=int,
                    method=partial(random.poisson, lam=3.14)
                ),
        },
        lambda key, shape: tree_map(
            lambda x: x.astype(float), 6. +
            random_draw(key, shape, int, partial(random.poisson, lam=3.14))
        ),
    ), (
        jft.VariableCovarianceGaussian,
        {
            "data": partial(random_draw, dtype=float, method=random.normal),
        },
        lambda key, shape: jft.Vector(
            (
                random_draw(key, shape, float, random.normal), 3. + 1. / jax.
                tree_map(
                    jnp.exp, random_draw(key, shape, float, random.normal)
                )
            )
        ),
    ), (
        jft.VariableCovarianceStudentT,
        {
            "data": partial(random_draw, dtype=float, method=random.normal),
            "dof": partial(random_draw, dtype=float, method=random.exponential),
        },
        lambda key, shape: jft.Vector(
            (
                random_draw(key, shape, float, random.normal),
                tree_map(
                    jnp.exp, 3. + 1e-1 *
                    random_draw(key, shape, float, random.normal)
                )
            )
        ),
    )
)


def _assert_zero_point_estimate_residuals(residuals, point_estimates):
    if not point_estimates:
        return
    assert_array_equal(jft.sum(point_estimates * residuals**2), 0.)


@pmp("seed", (42, 43))
@pmp("shape", ((5, 12), (5, ), (1, 2, 3, 4)))
@pmp("ndim", (1, 2, 3))
def test_concatenate_zip(seed, shape, ndim):
    keys = random.split(random.PRNGKey(seed), ndim)
    zip_args = tuple(random.normal(k, shape=shape) for k in keys)
    assert_array_equal(
        concatenate_zip(*zip_args),
        np.concatenate(list(zip(*(list(el) for el in zip_args))))
    )


@pmp("seed", (12, 42))
@pmp(
    "shape",
    ((5, ), ((4, ), (2, ), (1, ), (1, )), ((2, [1, 1]), {
        'a': (3, 1)
    }))
)
@pmp("lh_init", LH_INIT)
@pmp("random_point_estimates", (True, False))
@pmp("sample_mode", ("linear_resample", "nonlinear_resample"))
def test_optimize_kl_sample_consistency(
    seed, shape, lh_init, random_point_estimates, sample_mode
):
    rtol = 1e-7
    atol = 0.0
    aallclose = partial(assert_allclose, rtol=rtol, atol=atol)

    lh_init_method, draw, latent_init = lh_init
    key = random.PRNGKey(seed)
    key, *subkeys = random.split(key, 1 + len(draw))
    init_kwargs = {
        k: method(key=sk, shape=shape)
        for (k, method), sk in zip(draw.items(), subkeys)
    }
    lh = lh_init_method(**init_kwargs)
    key, sk = random.split(key)
    pos = latent_init(sk, shape=shape)
    jft.tree_math.assert_arithmetics(pos)

    if sample_mode == "nonlinear_resample":
        try:
            lh.transformation(pos)
        except NotImplementedError:
            pytest.skip("no transformation rule implemented yet")

    key, sk = random.split(key)
    delta = 1e-3
    absdelta = delta * jft.size(pos)

    draw_linear_kwargs = dict(
        cg_name="SL",
        cg_kwargs=dict(miniter=2, absdelta=absdelta / 10., maxiter=100),
    )
    minimize_kwargs = dict(
        name="SN",
        xtol=delta,
        cg_kwargs=dict(name=None, miniter=2),
        maxiter=5 if sample_mode.lower() == "nonlinear_resample" else 0,
    )
    key, sk = random.split(key)
    point_estimates = ()
    if random_point_estimates:
        prob = 0.5
        pe, pe_td = jax.tree_util.tree_flatten(pos)
        pe = list(random.bernoulli(sk, prob, shape=(len(pe), )))
        pe[0] = False  # Ensure at least one variable is not a point-estimate
        pe = tuple(map(bool, pe))
        point_estimates = jax.tree_util.tree_unflatten(pe_td, pe)

    residual_draw, _ = jft.draw_residual(
        lh,
        pos,
        sk,
        point_estimates=point_estimates,
        minimize_kwargs=minimize_kwargs,
        **draw_linear_kwargs,
    )
    _assert_zero_point_estimate_residuals(residual_draw, point_estimates)
    residual_diy_l1, _ = jft.draw_linear_residual(
        lh, pos, sk, point_estimates=point_estimates, **draw_linear_kwargs
    )
    residual_diy = jft.stack((residual_diy_l1, -residual_diy_l1))
    _assert_zero_point_estimate_residuals(residual_diy, point_estimates)
    if sample_mode.lower() == "linear":
        tree_map(aallclose, residual_draw, residual_diy)

    residual_diy_n1, _ = jft.nonlinearly_update_residual(
        lh,
        pos,
        residual_diy_l1,
        metric_sample_key=sk,
        metric_sample_sign=+1,
        point_estimates=point_estimates,
        minimize_kwargs=minimize_kwargs,
        jit=False,
    )
    residual_diy_n2, _ = jft.nonlinearly_update_residual(
        lh,
        pos,
        -residual_diy_l1,
        metric_sample_key=sk,
        metric_sample_sign=-1,
        point_estimates=point_estimates,
        minimize_kwargs=minimize_kwargs,
        jit=False,
    )
    residual_diy = jft.stack((residual_diy_n1, residual_diy_n2))
    _assert_zero_point_estimate_residuals(residual_diy, point_estimates)
    tree_map(aallclose, residual_draw, residual_diy)

    key, sk = random.split(key)
    samples_opt, _ = jft.optimize_kl(
        lh,
        pos,
        key=sk,
        n_total_iterations=1,
        n_samples=1,
        point_estimates=point_estimates,
        draw_linear_kwargs=draw_linear_kwargs,
        nonlinearly_update_kwargs=dict(minimize_kwargs=minimize_kwargs),
        kl_kwargs=dict(minimize_kwargs=dict(name="M", maxiter=0)),
        sample_mode=sample_mode,
        kl_jit=False,
        residual_jit=False,
        odir=None,
    )
    _assert_zero_point_estimate_residuals(samples_opt._samples, point_estimates)
    residual_draw, _ = jft.draw_residual(
        lh,
        pos,
        samples_opt.keys[0],
        point_estimates=point_estimates,
        minimize_kwargs=minimize_kwargs,
        **draw_linear_kwargs,
    )
    _assert_zero_point_estimate_residuals(residual_draw, point_estimates)
    tree_map(aallclose, samples_opt._samples, residual_draw)


if __name__ == "__main__":
    test_concatenate_zip(1, (5, 12), 3)
    test_optimize_kl_sample_consistency(
        1,
        ((2, [1, 1]), {
            'a': (3, 1)
        }),
        LH_INIT[1],
        random_point_estimates=True,
        sample_mode="nonlinear_resample",
    )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from .structured_domain import StructuredDomain


class GLSpace(StructuredDomain):
    """Represents a 2-sphere with Gauss-Legendre pixelization.

    Its harmonic partner domain is the
    :class:`~nifty8.domains.lm_space.LMSpace`.

    Parameters
    ----------
    nlat : int
        Number of latitudinal bins (or rings) that are used for this
        pixelization.
    nlon : int, optional
        Number of longitudinal bins that are used for this pixelization.
        Default value is 2*nlat - 1.
    """

    _needed_for_hash = ["_nlat", "_nlon"]

    def __init__(self, nlat, nlon=None):
        self._nlat = int(nlat)
        if self._nlat < 1:
            raise ValueError("nlat must be a positive number.")
        if nlon is None:
            self._nlon = 2*self._nlat - 1
        else:
            self._nlon = int(nlon)
            if self._nlon < 1:
                raise ValueError("nlon must be a positive number.")
        self._dvol = None

    def __repr__(self):
        return "GLSpace(nlat={}, nlon={})".format(self.nlat, self.nlon)

    @property
    def harmonic(self):
        return False

    @property
    def shape(self):
        return (int(self.nlat * self.nlon),)

    @property
    def size(self):
        return int(self.nlat * self.nlon)

    @property
    def scalar_dvol(self):
        return None

    # MR FIXME: this is potentially wasteful, since the return array is
    #           blown up by a factor of self.nlon
    @property
    def dvol(self):
        from ducc0.misc import GL_weights
        if self._dvol is None:
            self._dvol = GL_weights(self.nlat, self.nlon)
        return np.repeat(self._dvol, self.nlon)

    @property
    def total_volume(self):
        return 4*np.pi

    @property
    def nlat(self):
        """int : number of rings in this domain"""
        return self._nlat

    @property
    def nlon(self):
        """int : pixels per ring in this domain"""
        return self._nlon

    def get_default_codomain(self):
        """Returns a :class:`LMSpace` object, which is capable of storing an
        accurate representation of data residing on `self` (if this data is
        band-limited).

        Returns
        -------
        LMSpace
            The partner domain
        """
        from ..domains.lm_space import LMSpace
        mmax = self._nlon//2
        lmax = max(mmax, self._nlat-1)
        return LMSpace(lmax=lmax, mmax=mmax)

    def check_codomain(self, codomain):
        """Raises `TypeError` if `codomain` is not a matching partner domain
        for `self`.

        Notes
        -----
        This function only checks whether `codomain` is of type
        :class:`LMSpace`.
        """
        from ..domains.lm_space import LMSpace
        if not isinstance(codomain, LMSpace):
            raise TypeError("codomain must be a LMSpace.")



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2023 Vincent Eberle
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import h5py
import astropy.io.fits as ast
from numpy.testing import assert_array_equal

import nifty8 as ift


def test_optimize_kl_operator_output():
    sam_list = []
    dom = ift.RGSpace([32, 32])
    for i in range(10):
        tmp = ift.from_random(dom)
        sam_list.append(tmp)
    samples = ift.SampleList(sam_list)
    fname = "test_consistency"
    samples.save_to_fits(file_name_base=fname,
                         op=None,
                         samples=True,
                         mean=True,
                         std=True,
                         overwrite=True)
    samples.save_to_hdf5(file_name=fname+'.hdf5',
                         op=None,
                         samples=True,
                         mean=True,
                         std=True,
                         overwrite=True)

    mean, var = samples.sample_stat()
    mean = mean.val
    std = var.sqrt().val

    with ast.open(fname+"_mean.fits") as f:
        mean_fits = f[0].data
    with ast.open(fname+"_std.fits") as f:
        std_fits = f[0].data
    with h5py.File(fname+".hdf5", "r") as g:
        mean_hdf5 = g["stats"]["mean"][:]
        std_hdf5 = g["stats"]["standard deviation"][:]

    assert_array_equal(mean, mean_fits.T)
    assert_array_equal(std, std_fits.T)
    assert_array_equal(mean, mean_hdf5)
    assert_array_equal(std, std_hdf5)


def test_optimize_kl_domain_expansion():
    import numpy as np
    from tempfile import TemporaryDirectory

    dom = ift.RGSpace(10)

    R1 = ift.ScalingOperator(dom, 2.).ducktape('xi_r1')
    R2 = ift.ScalingOperator(dom, 3.).ducktape('xi_r2')
    N = ift.ScalingOperator(dom, 0.1**2, sampling_dtype=np.float64)

    data = R1(ift.from_random(R1.domain)) + N.draw_sample()
    lh_fn = ift.GaussianEnergy(data=data, inverse_covariance=N.inverse)

    sl = ift.optimize_kl(
        likelihood_energy=lambda i: lh_fn @ R1 if i == 0 else lh_fn @ (R1 + R2),
        total_iterations=2,
        n_samples=0,
        kl_minimizer=ift.NewtonCG(ift.AbsDeltaEnergyController(0.5, iteration_limit=1, name='Newton')),
        sampling_iteration_controller=None,
        nonlinear_sampling_minimizer=None,
        output_directory=TemporaryDirectory().name,
        plot_minisanity_history=True,
    )




# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_equal, assert_raises

from ..common import setup_function, teardown_function

pmp = pytest.mark.parametrize
# [lmax, expected]
CONSTRUCTOR_CONFIGS = [[
    5, None, {
        'lmax': 5,
        'mmax': 5,
        'shape': (36,),
        'harmonic': True,
        'size': 36,
    }
], [
    7, 4, {
        'lmax': 7,
        'mmax': 4,
        'shape': (52,),
        'harmonic': True,
        'size': 52,
    }
], [-1, 28, {
    'error': ValueError
}]]


def _k_length_array_helper(index_arr, lmax):
    if index_arr <= lmax:
        index_half = index_arr
    else:
        if (index_arr - lmax) % 2 == 0:
            index_half = (index_arr + lmax)//2
        else:
            index_half = (index_arr + lmax + 1)//2

    m = np.ceil(((2*lmax + 1) - np.sqrt((2*lmax + 1)**2 - 8 *
                                        (index_half - lmax)))/2).astype(int)

    return index_half - m*(2*lmax + 1 - m)//2


def get_k_length_array_configs():
    da_0 = [_k_length_array_helper(idx, 5) for idx in np.arange(36)]
    return [[5, da_0]]


@pmp('attribute', ['lmax', 'mmax', 'size'])
def test_property_ret_type(attribute):
    ift.myassert(isinstance(getattr(ift.LMSpace(7, 5), attribute), int))


@pmp('lmax, mmax, expected', CONSTRUCTOR_CONFIGS)
def test_constructor(lmax, mmax, expected):
    if 'error' in expected:
        with assert_raises(expected['error']):
            ift.LMSpace(lmax, mmax)
    else:
        for key, value in expected.items():
            assert_equal(getattr(ift.LMSpace(lmax, mmax), key), value)


def test_dvol():
    assert_allclose(ift.LMSpace(5).dvol, 1.)


@pmp('lmax, expected', get_k_length_array_configs())
def test_k_length_array(lmax, expected):
    assert_allclose(ift.LMSpace(lmax).get_k_length_array().val, expected)



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from functools import partial
import sys

import pytest

pytest.importorskip("jax")

import jax
from jax import random
import jax.numpy as jnp
from jax.tree_util import Partial
import numpy as np
from numpy.testing import assert_allclose
from scipy.spatial import distance_matrix

import nifty8.re as jft
from nifty8.re import refine
from nifty8.re.refine import charted_refine, healpix_refine

try:
    import healpy
except ImportError:
    healpy = None

pmp = pytest.mark.parametrize


def matern_kernel(distance, scale, cutoff, dof):
    from jax.scipy.special import gammaln
    from scipy.special import kv

    reg_dist = jnp.sqrt(2 * dof) * distance / cutoff
    return scale**2 * 2**(1 - dof) / jnp.exp(
        gammaln(dof)
    ) * (reg_dist)**dof * kv(dof, reg_dist)


scale, cutoff, dof = 1., 80., 3 / 2

x = jnp.logspace(-6, 11, base=jnp.e, num=int(1e+5))
y = matern_kernel(x, scale, cutoff, dof)
y = jnp.nan_to_num(y, nan=0.)
kernel = Partial(jnp.interp, xp=x, fp=y)
inv_kernel = Partial(jnp.interp, xp=y, fp=x)


@pmp("dist", (10., 1e+3))
def test_refinement_matrices_1d(dist, kernel=kernel):
    cov_from_loc = refine.util.get_cov_from_loc(kernel=kernel)

    coarse_coord = dist * jnp.array([0., 1., 2.])
    fine_coord = coarse_coord[tuple(
        jnp.array(coarse_coord.shape) // 2
    )] + (jnp.diff(coarse_coord) / jnp.array([-4., 4.]))
    cov_ff = cov_from_loc(fine_coord, fine_coord)
    cov_fc = cov_from_loc(fine_coord, coarse_coord)
    cov_cc_inv = jnp.linalg.inv(cov_from_loc(coarse_coord, coarse_coord))

    fine_kernel = cov_ff - cov_fc @ cov_cc_inv @ cov_fc.T
    fine_kernel_sqrt_diy = jnp.linalg.cholesky(fine_kernel)
    olf_diy = cov_fc @ cov_cc_inv

    cc = jft.CoordinateChart(
        shape0=(12, ) * np.size(dist),
        distances=dist,
        depth=0,
        _coarse_size=3,
        _fine_size=2
    )
    olf, ks = jft.RefinementField(cc).matrices_at(
        level=0, pixel_index=(0, ) * np.size(dist), kernel=kernel
    )

    assert_allclose(olf, olf_diy)
    assert_allclose(ks, fine_kernel_sqrt_diy)


@pmp("seed", (12, 45))
@pmp("dist", (10., 1e+3))
def test_refinement_1d(seed, dist, kernel=kernel):
    rng = np.random.default_rng(seed)

    refs = (
        charted_refine.refine_conv, charted_refine.refine_conv_general,
        charted_refine.refine_loop, charted_refine.refine_vmap,
        charted_refine.refine_loop, charted_refine.refine_slice
    )
    cov_from_loc = refine.util.get_cov_from_loc(kernel=kernel)
    cc = jft.CoordinateChart(
        shape0=(12, ) * np.size(dist),
        distances=dist,
        depth=0,
        _coarse_size=3,
        _fine_size=2
    )
    olf, ks = jft.RefinementField(cc).matrices_at(
        level=0, pixel_index=(0, ) * np.size(dist), kernel=kernel
    )

    main_coord = jnp.linspace(0., 1000., 50)
    cov_sqrt = jnp.linalg.cholesky(cov_from_loc(main_coord, main_coord))
    lvl0 = cov_sqrt @ rng.normal(size=main_coord.shape)
    lvl1_exc = rng.normal(size=(2 * (lvl0.size - 2), ))

    fine_reference = charted_refine.refine(lvl0, lvl1_exc, olf, ks)
    eps = jnp.finfo(lvl0.dtype.type).eps
    aallclose = partial(
        assert_allclose, desired=fine_reference, rtol=6 * eps, atol=120 * eps
    )
    for ref in refs:
        print(f"testing {ref.__name__}", file=sys.stderr)
        aallclose(ref(lvl0, lvl1_exc, olf, ks))


@pmp("seed", (12, ))
@pmp("dist", (60., 1e+3, (40., 90.), (1e+2, 1e+3, 1e+4)))
@pmp("_coarse_size", (3, 5))
@pmp("_fine_size", (2, 4))
@pmp("_fine_strategy", ("jump", "extend"))
def test_refinement_nd_cross_consistency(
    seed, dist, _coarse_size, _fine_size, _fine_strategy, kernel=kernel
):
    ndim = len(dist) if hasattr(dist, "__len__") else 1
    min_shape = (12, ) * ndim
    depth = 1
    refs = (charted_refine.refine_conv_general, charted_refine.refine_slice)
    kwargs = {
        "_coarse_size": _coarse_size,
        "_fine_size": _fine_size,
        "_fine_strategy": _fine_strategy
    }

    chart = jft.CoordinateChart(
        min_shape, depth=depth, distances=dist, **kwargs
    )
    rfm = jft.RefinementField(chart).matrices(kernel)
    xi = jft.random_like(
        random.PRNGKey(seed),
        jft.RefinementField(chart).domain
    )

    cf = partial(jft.RefinementField.apply, chart=chart, kernel=rfm)
    fine_reference = cf(xi)
    eps = jnp.finfo(fine_reference.dtype.type).eps
    aallclose = partial(
        assert_allclose, desired=fine_reference, rtol=6 * eps, atol=60 * eps
    )
    for ref in refs:
        print(f"testing {ref.__name__}", file=sys.stderr)
        aallclose(cf(xi, _refine=ref))


@pmp("dist", (60., 1e+3, (40., 90.), (1e+2, 1e+3, 1e+4)))
def test_refinement_fine_strategy_basic_consistency(dist, kernel=kernel):
    shape0 = (12, ) * len(dist) if isinstance(dist, tuple) else (12, )

    cc_e = jft.CoordinateChart(
        shape0=shape0,
        distances=dist,
        depth=0,
        _fine_strategy="extend",
        _coarse_size=3,
        _fine_size=2
    )
    olf_e, ks_e = jft.RefinementField(cc_e).matrices_at(
        level=0, pixel_index=(0, ) * np.size(dist), kernel=kernel
    )
    cc_j = jft.CoordinateChart(
        shape0=shape0,
        distances=dist,
        depth=0,
        _fine_strategy="jump",
        _coarse_size=3,
        _fine_size=2
    )
    olf_j, ks_j = jft.RefinementField(cc_j).matrices_at(
        level=0, pixel_index=(0, ) * np.size(dist), kernel=kernel
    )

    assert_allclose(olf_j, olf_e, rtol=1e-13, atol=0.)
    assert_allclose(ks_j, ks_e, rtol=1e-13, atol=0.)

    depth = 2
    rm_j = jft.RefinementField(cc_j).matrices(kernel=kernel, depth=depth)
    rm_e = jft.RefinementField(cc_e).matrices(kernel=kernel, depth=depth)

    assert_allclose(rm_j.filter, rm_e.filter, rtol=1e-13, atol=0.)
    assert_allclose(
        rm_j.propagator_sqrt, rm_e.propagator_sqrt, rtol=1e-13, atol=0.
    )
    assert_allclose(rm_j.cov_sqrt0, rm_e.cov_sqrt0, rtol=1e-13, atol=0.)


@pmp("dist", (60., 1e+3, (40., 90.), (1e+2, 1e+3, 1e+4)))
@pmp("_coarse_size", (3, 5))
@pmp("_fine_size", (2, 4))
@pmp("_fine_strategy", ("jump", "extend"))
def test_refinement_covariance(
    dist, _coarse_size, _fine_size, _fine_strategy, kernel=kernel
):
    distances0 = np.atleast_1d(dist)
    ndim = len(distances0)

    cf = jft.RefinementField(
        shape0=(_coarse_size, ) * ndim,
        depth=1,
        _coarse_size=_coarse_size,
        _fine_size=_fine_size,
        _fine_strategy=_fine_strategy,
        distances0=distances0,
        kernel=kernel
    )
    exc_shp = [
        jft.ShapeWithDtype((_coarse_size, ) * ndim),
        jft.ShapeWithDtype((_fine_size, ) * ndim)
    ]
    cf_shp = jax.eval_shape(cf, exc_shp)
    assert cf_shp.shape == (_fine_size, ) * ndim

    probe = jnp.zeros(cf_shp.shape)
    indices = np.indices(cf_shp.shape).reshape(ndim, -1)
    # Work around jax.linear_transpose NotImplementedError
    _, cf_T = jax.vjp(cf, jft.zeros_like(exc_shp))
    cf_cf_T = lambda x: cf(*cf_T(x))
    cov_empirical = jax.vmap(
        lambda idx: cf_cf_T(probe.at[tuple(idx)].set(1.)).ravel(),
        in_axes=1,
        out_axes=-1
    )(indices)

    pos = np.mgrid[tuple(slice(s) for s in cf_shp.shape)].astype(float)
    if _fine_strategy == "jump":
        pos *= distances0.reshape((-1, ) + (1, ) * ndim) / _fine_size
    elif _fine_strategy == "extend":
        pos *= distances0.reshape((-1, ) + (1, ) * ndim) / 2
    else:
        raise AssertionError(f"invalid `_fine_strategy`; {_fine_strategy}")
    pos = jnp.moveaxis(pos, 0, -1)
    p = pos.reshape(-1, ndim)
    dist_mat = distance_matrix(p, p)
    cov_truth = kernel(dist_mat)

    assert_allclose(cov_empirical, cov_truth, rtol=1e-14, atol=1e-15)


@pmp("seed", (12, 42, 43, 45))
@pmp("n_dim", (1, 2, 3, 4, 5))
def test_refinement_nd_shape(seed, n_dim, kernel=kernel):
    rng = np.random.default_rng(seed)

    distances = np.exp(rng.normal(size=(n_dim, )))
    cov_from_loc = refine.util.get_cov_from_loc(kernel=kernel)
    cc = jft.CoordinateChart(
        shape0=(12, ) * n_dim,
        distances=distances,
        depth=0,
        _coarse_size=3,
        _fine_size=2
    )
    olf, ks = jft.RefinementField(cc).matrices_at(
        level=0,
        pixel_index=(0, ) * n_dim,
        kernel=kernel,
        coerce_fine_kernel=False
    )

    shp_i = 5
    gc = distances.reshape(n_dim, 1) * jnp.linspace(0., 1000., shp_i)
    gc = jnp.stack(jnp.meshgrid(*gc, indexing="ij"), axis=-1).reshape(-1, n_dim)
    cov_sqrt = jnp.linalg.cholesky(cov_from_loc(gc, gc))
    lvl0 = (cov_sqrt @ rng.normal(size=gc.shape[0])).reshape((shp_i, ) * n_dim)
    lvl1_exc = rng.normal(size=tuple(n - 2 for n in lvl0.shape) + (2**n_dim, ))

    fine_reference = charted_refine.refine(lvl0, lvl1_exc, olf, ks)
    assert fine_reference.shape == tuple((2 * (shp_i - 2), ) * n_dim)


@pmp("dist", (60., 1e+3, (40., 90.)))
@pmp("_coarse_size", (3, 5))
@pmp("_fine_size", (2, 4))
@pmp("_fine_strategy", ("jump", "extend"))
def test_chart_refinement_regular_irregular_matrices_consistency(
    dist, _coarse_size, _fine_size, _fine_strategy, kernel=kernel
):
    depth = 2
    distances = np.atleast_1d(dist)
    ndim = distances.size
    kwargs = {
        "_coarse_size": _coarse_size,
        "_fine_size": _fine_size,
        "_fine_strategy": _fine_strategy
    }

    cc = jft.CoordinateChart(
        (12, ) * ndim, depth=depth, distances=distances, **kwargs
    )
    refinement = jft.RefinementField(cc).matrices(kernel=kernel)

    cc_irreg = jft.CoordinateChart(
        shape0=cc.shape0,
        depth=depth,
        distances=distances,
        irregular_axes=tuple(range(ndim)),
        **kwargs
    )
    refinement_irreg = jft.RefinementField(cc_irreg).matrices(kernel=kernel)

    aallclose = partial(assert_allclose, rtol=1e-14, atol=1e-13)
    aallclose(refinement.cov_sqrt0, refinement_irreg.cov_sqrt0)

    for lvl in range(depth):
        olf, ks = refinement.filter[lvl], refinement.propagator_sqrt[lvl]
        olf_irreg, ks_irreg = refinement_irreg.filter[
            lvl], refinement_irreg.propagator_sqrt[lvl]

        olf_d = np.diff(
            olf_irreg.reshape((-1, ) + olf_irreg.shape[-2:]), axis=0
        )
        ks_d = np.diff(ks_irreg.reshape((-1, ) + ks_irreg.shape[-2:]), axis=0)
        aallclose(olf_d, 0.)
        aallclose(ks_d, 0.)
        aallclose(olf_irreg[(0, ) * ndim], olf.squeeze())
        aallclose(ks_irreg[(0, ) * ndim], ks.squeeze())


@pmp("seed", (12, ))
@pmp("dist", (60., 1e+3, (40., 90.), (1e+2, 1e+3, 1e+4)))
@pmp("_coarse_size", (3, 5))
@pmp("_fine_size", (2, 4))
@pmp("_fine_strategy", ("jump", "extend"))
@pmp(
    "_refine",
    (charted_refine.refine_conv_general, charted_refine.refine_slice)
)
def test_refinement_irregular_regular_consistency(
    seed,
    dist,
    _coarse_size,
    _fine_size,
    _fine_strategy,
    _refine,
    kernel=kernel
):
    depth = 1
    distances = np.atleast_1d(dist)
    ndim = distances.size
    kwargs = {
        "_coarse_size": _coarse_size,
        "_fine_size": _fine_size,
        "_fine_strategy": _fine_strategy
    }

    cc = jft.RefinementField(
        shape0=(2 * _coarse_size, ) * ndim,
        depth=depth,
        distances=distances,
        **kwargs
    )
    refinement = cc.matrices(kernel=kernel)

    cc_irreg = jft.RefinementField(
        shape0=cc.chart.shape0,
        depth=depth,
        distances=distances,
        irregular_axes=tuple(range(ndim)),
        **kwargs
    )
    refinement_irreg = cc_irreg.matrices(kernel=kernel)

    rng = np.random.default_rng(seed)
    exc_swd = cc.domain[-1]
    fn1 = rng.normal(size=cc.chart.shape_at(depth - 1))
    exc = rng.normal(size=exc_swd.shape)

    refined = _refine(
        fn1, exc, refinement.filter[-1], refinement.propagator_sqrt[-1],
        **kwargs
    )
    refined_irreg = _refine(
        fn1, exc, refinement_irreg.filter[-1],
        refinement_irreg.propagator_sqrt[-1], **kwargs
    )
    assert_allclose(refined_irreg, refined, rtol=1e-14, atol=1e-13)


@pytest.mark.skipif(
    healpy is None, reason="requires the software library `healpy`"
)
@pmp("nside", (1, 2, 16))
@pmp("nest", (True, False))
def test_healpix_refinement_neigbor_uniqueness(nside, nest):
    nbr = healpix_refine.get_all_1st_hp_nbrs_idx(nside, nest)
    n_non_uniq = np.sum(np.diff(np.sort(nbr, axis=1), axis=1) == 0, axis=1)
    np.testing.assert_array_equal(nbr == -1, False)
    np.testing.assert_equal(n_non_uniq, 0)


if __name__ == "__main__":
    test_refinement_matrices_1d(5.)
    test_refinement_1d(42, 10.)
    test_healpix_refinement_neigbor_uniqueness(1, True)
    test_healpix_refinement_neigbor_uniqueness(2, True)



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

import pytest

pytest.importorskip("jax")

from functools import partial
import sys
from jax import random
import jax.numpy as jnp
import numpy as np
from numpy.testing import assert_allclose
from scipy.spatial import distance_matrix

import nifty8.re as jft

pmp = pytest.mark.parametrize


def matern_kernel(distance, scale, cutoff, dof):
    from jax.scipy.special import gammaln
    from scipy.special import kv

    reg_dist = jnp.sqrt(2 * dof) * distance / cutoff
    cov = scale**2 * 2**(1 - dof) / jnp.exp(
        gammaln(dof)
    ) * (reg_dist)**dof * kv(dof, reg_dist)
    # NOTE, this is not safe for differentiating because `cov` still may
    # contain NaNs
    return jnp.where(distance < 1e-8 * cutoff, scale**2, cov)


from operator import matmul


@pmp("seed", tuple(range(12, 44, 5)))
@pmp("shape0", (128, 64))
def test_lanczos_tridiag(seed, shape0):
    rng = np.random.default_rng(seed)
    rng_key = random.PRNGKey(rng.integers(12, 42))

    m = rng.normal(size=(shape0, ) * 2)
    m = m @ m.T  # ensure positive-definiteness

    v = random.rademacher(rng_key, (shape0, ), float)
    tridiag, vecs = jft.lanczos.lanczos_tridiag(partial(matmul, m), v, shape0)
    m_est = vecs.T @ tridiag @ vecs

    assert_allclose(m_est, m, atol=1e-13, rtol=1e-13)


@pmp("seed", tuple(range(12, 44, 5)))
@pmp("shape0", (128, 64))
def test_stochastic_lq_logdet(seed, shape0, lq_order=15, n_lq_samples=10):
    rng = np.random.default_rng(seed)
    rng_key = random.PRNGKey(rng.integers(12, 42))

    c = np.exp(3 + rng.normal())
    s = np.exp(rng.normal())

    p = np.logspace(np.log(0.1 * c), np.log(1e+2 * c), num=shape0 - 1)
    p = np.concatenate(([0], p)).reshape(-1, 1)

    m = jnp.asarray(
        matern_kernel(distance_matrix(p, p), cutoff=c, scale=s, dof=2.5)
    )

    _, logdet = jnp.linalg.slogdet(m)
    logdet_est = jft.stochastic_lq_logdet(m, lq_order, n_lq_samples, rng_key)
    assert_allclose(logdet_est, logdet, rtol=2., atol=20.)
    print(f"{logdet=} :: {logdet_est=}", file=sys.stderr)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest
from numpy.testing import assert_allclose

from ..common import setup_function, teardown_function

pmp = pytest.mark.parametrize
spaces = [
    ift.RGSpace(4),
    ift.RGSpace((4, 4), (0.123, 0.4)),
    ift.HPSpace(8),
    ift.GLSpace(4)
]


def _stats(op, samples):
    sc = ift.StatCalculator()
    for s in samples:
        sc.add(op(s.extract(op.domain)))
    return sc.mean.val, sc.var.ptw("sqrt").val


def _rand():
    return ift.random.current_rng().normal()


def _posrand():
    return np.exp(_rand())


@pmp('dofdex', [[0, 0], [0, 1]])
@pmp('seed', [12, 3])
def testDistributor(dofdex, seed):
    with ift.random.Context(seed):
        dom = ift.RGSpace(3)
        N_copies = max(dofdex) + 1
        distributed_target = ift.makeDomain(
            (ift.UnstructuredDomain(len(dofdex)), dom))
        target = ift.makeDomain((ift.UnstructuredDomain(N_copies), dom))
        op = ift.library.correlated_fields._Distributor(
            dofdex, target, distributed_target)
        ift.extra.check_linear_operator(op)


@pmp('total_N', [0, 1, 2])
@pmp('offset_std', [None, (1, 1)])
@pmp('asperity', [None, (1, 1)])
@pmp('flexibility', [None, (1, 1)])
@pmp('ind', [None, 1])
@pmp('matern', [True, False])
def test_init(total_N, offset_std, asperity, flexibility, ind, matern):
    if flexibility is None and asperity is not None:
        pytest.skip()
    cfg = 1, 1
    for dofdex in ([None], [None, [0]], [None, [0, 0], [0, 1], [1, 1]])[total_N]:
        cfm = ift.CorrelatedFieldMaker('', total_N)
        cfm.add_fluctuations(ift.RGSpace(4), cfg, flexibility, asperity, (-2, 0.1))
        if not offset_std is None:
            if matern:
                if total_N == 0:
                    cfm.add_fluctuations_matern(ift.RGSpace(4), *(3*[cfg]))
                else:
                    with pytest.raises(NotImplementedError):
                        cfm.add_fluctuations_matern(ift.RGSpace(4), *(3*[cfg]))
            else:
                cfm.add_fluctuations(ift.RGSpace(4), *(4*[cfg]), index=ind)
        cfm.set_amplitude_total_offset(0, offset_std, dofdex=dofdex)
        cfm.finalize(prior_info=0)


@pmp('sspace', spaces)
@pmp('asperity', [None, (1, 1)])
@pmp('flexibility', [None, (1, 1)])
@pmp('matern', [True, False])
def test_unit_zero_mode(sspace, asperity, flexibility, matern):
    if flexibility is None and asperity is not None:
        pytest.skip()
    cfg = 1, 1
    cfm = ift.CorrelatedFieldMaker('')
    if matern:
        cfm.add_fluctuations_matern(sspace, *(3 * [cfg]))
    else:
        cfm.add_fluctuations(sspace, *(4 * [cfg]))
    cfm.set_amplitude_total_offset(0, 1.)
    cf = cfm.finalize(prior_info=0)

    r = ift.from_random(cf.domain).to_dict()
    r_xi = np.copy(r["xi"].val)
    r_xi[0] = 1.
    r["xi"] = ift.Field(r["xi"].domain, r_xi)
    r = ift.MultiField.from_dict(r)

    cf_r = cf(r)
    rtol = 1e-7
    if isinstance(sspace, (ift.HPSpace, ift.GLSpace)):
        rtol = 1e-2
    assert_allclose(cf_r.s_integrate(), sspace.total_volume, rtol=rtol)

@pmp('sspace', spaces)
@pmp('asperity', [None, (1, 1)])
@pmp('flexibility', [None, (1, 1)])
@pmp('matern', [True, False])
def test_constant_zero_mode(sspace, asperity, flexibility, matern):
    if flexibility is None and asperity is not None:
        pytest.skip()
    cfg = 1, 1
    cfm = ift.CorrelatedFieldMaker('')
    if matern:
        cfm.add_fluctuations_matern(sspace, *(3 * [cfg]))
    else:
        fl = (1., 0.5)
        ll = (-4, 1)
        cfm.add_fluctuations(sspace, fl, flexibility, asperity, ll)
        
        cf_simple = ift.SimpleCorrelatedField(sspace, 0, None, fl, flexibility, asperity, ll)
    cfm.set_amplitude_total_offset(0, None)
    cf = cfm.finalize(prior_info=0)

    r = ift.from_random(cf.domain)
    cf_r = cf(r)
    atol = 1e-8
    if isinstance(sspace, (ift.HPSpace, ift.GLSpace)):
        atol = 1e-2
        if matern:
            atol = 1e-1
    assert_allclose(cf_r.s_integrate(), 0., atol=atol)
    if not matern:
        cf_r_simple = cf_simple(r)
        assert_allclose(cf_r_simple.s_integrate(), 0., atol=atol)

@pmp('sspace', spaces)
@pmp('N', [0, 2])
def testAmplitudesInvariants(sspace, N):
    fsspace = ift.RGSpace((12,), (0.4,))
    dofdex1, dofdex2, dofdex3 = None, None, None
    if N == 2:
        dofdex1, dofdex2, dofdex3 = [0, 0], [1, 0], [1, 1]

    astds = 0.2, 1.2
    offset_std_mean = 1.3
    fa = ift.CorrelatedFieldMaker('', N)
    fa.add_fluctuations(sspace, (astds[0], 1e-2), (1.1, 2.), (2.1, .5), (-2, 1.),
                        'spatial', dofdex=dofdex2)
    fa.add_fluctuations(fsspace, (astds[1], 1e-2), (3.1, 1.), (.5, .1), (-4, 1.),
                        'freq', dofdex=dofdex3)
    fa.set_amplitude_total_offset(1.2, (offset_std_mean, 1e-2), dofdex=dofdex1)
    op = fa.finalize(prior_info=0)

    samples = [ift.from_random(op.domain) for _ in range(100)]
    tot_flm, _ = _stats(fa.total_fluctuation, samples)
    offset_amp_std, _ = _stats(fa.amplitude_total_offset, samples)
    intergated_fluct_std0, _ = _stats(fa.average_fluctuation(0), samples)
    intergated_fluct_std1, _ = _stats(fa.average_fluctuation(1), samples)

    slice_fluct_std0, _ = _stats(fa.slice_fluctuation(0), samples)
    slice_fluct_std1, _ = _stats(fa.slice_fluctuation(1), samples)

    sams = [op(s) for s in samples]
    fluct_total = fa.total_fluctuation_realized(sams)
    fluct_space = fa.average_fluctuation_realized(sams, 0)
    fluct_freq = fa.average_fluctuation_realized(sams, 1)
    zm_std_mean = fa.offset_amplitude_realized(sams)
    sl_fluct_space = fa.slice_fluctuation_realized(sams, 0)
    sl_fluct_freq = fa.slice_fluctuation_realized(sams, 1)

    assert_allclose(offset_amp_std, zm_std_mean, rtol=0.5)
    assert_allclose(intergated_fluct_std0, fluct_space, rtol=0.5)
    assert_allclose(intergated_fluct_std1, fluct_freq, rtol=0.5)
    assert_allclose(tot_flm, fluct_total, rtol=0.5)
    assert_allclose(slice_fluct_std0, sl_fluct_space, rtol=0.5)
    assert_allclose(slice_fluct_std1, sl_fluct_freq, rtol=0.5)

    fa = ift.CorrelatedFieldMaker('', N)
    fa.set_amplitude_total_offset(0., (offset_std_mean, .1), dofdex=dofdex1)
    fa.add_fluctuations(fsspace, (astds[1], 1.), (3.1, 1.), (.5, .1), (-4, 1.), 'freq',
                        dofdex=dofdex3)
    m = 3.
    x = fa.moment_slice_to_average(m)
    fa.add_fluctuations(sspace, (x, 1.5), (1.1, 2.), (2.1, .5), (-2, 1.), 'spatial', 0,
                        dofdex=dofdex2)
    op = fa.finalize(prior_info=0)
    em, estd = _stats(fa.slice_fluctuation(0), samples)

    assert_allclose(m, em, rtol=0.5)
    assert op.target[-2] == sspace
    assert op.target[-1] == fsspace

    for ampl in fa.get_normalized_amplitudes():
        ift.extra.check_operator(ampl, 0.1*ift.from_random(ampl.domain), ntries=10)
    ift.extra.check_operator(op, 0.1*ift.from_random(op.domain), ntries=10)


@pmp('seed', [42, 31])
@pmp('domain', spaces)
@pmp('without', (('offset_std', ), ('asperity', ), ('flexibility', ), ('flexibility', 'asperity')))
def test_complicated_vs_simple(seed, domain, without):
    with ift.random.Context(seed):
        offset_mean = _rand()
        fluctuations = _posrand(), _posrand()
        if "flexibility" in without:
            flexibility = None
        else:
            flexibility = _posrand(), _posrand()
        if "asperity" in without:
            asperity = None
        else:
            asperity = _posrand(), _posrand()
        if "offset_std" in without:
            offset_std = None
        else:
            offset_std = _posrand(), _posrand()
        loglogavgslope = _posrand(), _posrand()
        prefix = 'foobar'
        hspace = domain.get_default_codomain()
        scf_args = (
            domain,
            offset_mean,
            offset_std,
            fluctuations,
            flexibility,
            asperity,
            loglogavgslope
        )
        add_fluct_args = (
            domain,
            fluctuations,
            flexibility,
            asperity,
            loglogavgslope
        )
        cfm = ift.CorrelatedFieldMaker(prefix)
        if asperity is not None and flexibility is None:
            with pytest.raises(ValueError):
                scf = ift.SimpleCorrelatedField(*scf_args, prefix=prefix,
                                                harmonic_partner=hspace)
            with pytest.raises(ValueError):
                cfm.add_fluctuations(*add_fluct_args, prefix='',
                                     harmonic_partner=hspace)
            return
        scf = ift.SimpleCorrelatedField(*scf_args, prefix=prefix,
                                        harmonic_partner=hspace)
        cfm.add_fluctuations(*add_fluct_args, prefix='',
                             harmonic_partner=hspace)
        cfm.set_amplitude_total_offset(offset_mean, offset_std)
        inp = ift.from_random(scf.domain)
        op1 = cfm.finalize(prior_info=0)
        assert scf.domain is op1.domain
        ift.extra.assert_allclose(scf(inp), op1(inp))
        ift.extra.check_operator(scf, inp, ntries=10)

        op1 = cfm.amplitude
        op0 = scf.amplitude
        assert op0.domain is op1.domain
        ift.extra.assert_allclose(op0.force(inp), op1.force(inp))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2019-2020 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

import nifty8 as ift


def convtest(test_signal, delta, func):
    domain = test_signal.domain

    # Create Convolution Operator
    conv_op = ift.FuncConvolutionOperator(domain, func)

    # Convolve, Adjoint-Convolve
    conv_signal = conv_op(test_signal)
    cac_signal = conv_op.adjoint_times(conv_signal)

    print(test_signal.integrate(), conv_signal.integrate(),
          cac_signal.integrate())

    # generate kernel image
    conv_delta = conv_op(delta)

    # Plot results
    plot = ift.Plot()
    plot.add(test_signal, title='Signal')
    plot.add(conv_signal, title='Signal Convolved')
    plot.add(cac_signal, title='Signal, Conv, Adj-Conv')
    plot.add(conv_delta, title='Kernel')
    plot.output()


def main():
    # Healpix test
    nside = 64
    npix = 12 * nside * nside

    domain = ift.HPSpace(nside)

    # Define test signal (some point sources)
    signal_vals = np.zeros(npix, dtype=np.float64)
    for i in range(0, npix, npix//12 + 27):
        signal_vals[i] = 500.

    signal = ift.makeField(domain, signal_vals)

    delta_vals = np.zeros(npix, dtype=np.float64)
    delta_vals[0] = 1.0
    delta = ift.makeField(domain, delta_vals)

    # Define kernel function
    def func(theta):
        ct = np.cos(theta)
        return 1. * np.logical_and(ct > 0.7, ct <= 0.8)

    convtest(signal, delta, func)

    domain = ift.RGSpace((100, 100))
    # Define test signal (some point sources)
    signal_vals = np.zeros(domain.shape, dtype=np.float64)
    signal_vals[35, 70] = 5000.
    signal_vals[45, 8] = 5000.
    signal = ift.makeField(domain, signal_vals)

    # Define delta signal, generate kernel image
    delta_vals = np.zeros(domain.shape, dtype=np.float64)
    delta_vals[0, 0] = 1.0
    delta = ift.makeField(domain, delta_vals)

    convtest(signal, delta, lambda d: 1. * np.logical_and(d > 0.1, d <= 0.2))


if __name__ == '__main__':
    main()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from glob import glob
from os import remove

import pytest


def list2fixture(lst):
    @pytest.fixture(params=lst)
    def myfixture(request):
        return request.param

    return myfixture


def setup_function():
    import nifty8 as ift
    ift.random.push_sseq_from_seed(42)
    comm, _, _, master = ift.utilities.get_MPI_params()
    if master:
        for ff in glob("*.pickle") + glob("*.png") + glob("*.h5"):
            remove(ff)
    if comm is not None:
        comm.Barrier()


def teardown_function():
    import nifty8 as ift
    ift.random.pop_sseq()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2022 Max-Planck-Society
# Author: Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest
from numpy.testing import assert_allclose
from scipy.stats import gamma, invgamma, norm

from ..common import list2fixture, setup_function, teardown_function

pmp = pytest.mark.parametrize
space = list2fixture([ift.GLSpace(15),
                      ift.RGSpace(64, distances=.789),
                      ift.RGSpace([32, 32], distances=.789)])
seed = list2fixture([4, 78, 23])


def testInterpolationAccuracy(space, seed):
    ift.random.push_sseq_from_seed(seed)
    pos = ift.from_random(space, 'normal')
    alpha = 1.5
    qs = [0.73, pos.ptw("exp").val]
    for q in qs:
        qfld = q
        if not np.isscalar(q):
            qfld = ift.makeField(space, q)
        op = ift.InverseGammaOperator(space, alpha, qfld)
        op1 = ift.LogInverseGammaOperator(space, alpha, qfld).exp()
        arr1 = op(pos).val
        arr2 = op1(pos).val
        arr0 = invgamma.ppf(norm.cdf(pos.val), alpha, scale=q)
        assert_allclose(arr0, arr1)
        assert_allclose(arr0, arr2)

        op2 = ift.GammaOperator(space, alpha=alpha, theta=qfld)
        arr1 = op2(pos).val
        arr0 = gamma.ppf(norm.cdf(pos.val), alpha, scale=q)
        assert_allclose(arr0, arr1)



# Copyright(C) 2022-2023 Gordian Edenhofer
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from functools import partial

import jax
from jax import lax
from jax.tree_util import tree_map
from jax import numpy as jnp


def _int_or_none(x):
    return isinstance(x, int) or x is None


def _fun_reord(_, mapped, *, fun, unmapped, unflatten, in_axes):
    un, mapped = list(unmapped), list(mapped)
    assert len(un) + len(mapped) == len(in_axes)
    args = tuple(un.pop(0) if a is None else mapped.pop(0) for a in in_axes)
    y = fun(*unflatten(args))
    return None, y


def _moveaxis(a, source, destination):
    # Ensure that arrays are never completely unnecessarily copied
    if source == destination:
        return a
    return jnp.moveaxis(a, source, destination)


def _generic_smap(fun, in_axes, out_axes, unroll, *x, _scan=lax.scan, **k):
    from jax.tree_util import tree_flatten, tree_map, tree_unflatten

    if k:
        raise TypeError("keyword arguments are not allowed in map")

    if isinstance(in_axes, int):
        in_axes = tree_map(lambda _: in_axes, x)
    elif isinstance(in_axes, tuple):
        if len(in_axes) != len(x):
            ve = f"`in_axes` {in_axes!r} and input {x!r} must be of same length"
            raise ValueError(ve)
        new_inax = []
        for el, i in zip(x, in_axes):
            new_inax.append(tree_map(lambda _: i, el) if _int_or_none(i) else i)
        in_axes = tuple(new_inax)
    else:
        te = f"`in_axes` must be an int or tuple of pytrees/int; got {in_axes!r}"
        raise TypeError(te)
    x, x_td = tree_flatten(x)
    in_axes, in_axes_td = tree_flatten(in_axes, is_leaf=_int_or_none)
    if in_axes_td != x_td:
        ve = f"`in_axes` {in_axes_td!r} incompatible with input `*x` {x_td!r}"
        raise ValueError(ve)

    unmapped = []
    mapped = []
    for i, el in zip(in_axes, x):
        if i is None:
            unmapped.append(el)
        elif isinstance(i, int):
            mapped.append(_moveaxis(el, i, 0))
        else:
            raise TypeError(f"expected `in_axes` index of type int; got {i!r}")

    fun_reord = partial(
        _fun_reord,
        fun=fun,
        unmapped=unmapped,
        unflatten=partial(tree_unflatten, x_td),
        in_axes=in_axes
    )
    _, y = _scan(fun_reord, None, mapped, unroll=unroll)

    if out_axes is None:
        out_axes, out_axes_td = tree_flatten(out_axes)
    if isinstance(out_axes, int):
        out_axes = tree_map(lambda el: out_axes if el is not None else el, y)
        out_axes, out_axes_td = tree_flatten(out_axes)
    else:
        out_axes, out_axes_td = tree_flatten(out_axes, is_leaf=_int_or_none)
    y, y_td = tree_flatten(y)
    if y is not None and out_axes_td != y_td:
        ve = f"`out_axes` {out_axes_td!r} incompatible with output {y_td!r}"
        raise ValueError(ve)
    out = []
    for i, el in zip(out_axes, y):
        if i is None:
            out.append(unmapped.pop(0))
        elif isinstance(i, int):
            out.append(_moveaxis(el, 0, i))
        else:
            raise TypeError(f"expected `out_axes` index of type int; got {i!r}")

    return tree_unflatten(y_td, out)


# The function over which to `scan` depends on the data. This leads to
# unnecessary recompiles. Ensure scan is compiled only once by compiling the
# whole data dependence.
_smap = jax.jit(
    _generic_smap,
    static_argnames=("fun", "in_axes", "out_axes", "unroll", "_scan")
)


def smap(fun, in_axes=0, out_axes=0, *, unroll=1):
    """Stupid/sequential map.

    Many of JAX's control flow logic reduces to a simple `jax.lax.scan`. This
    function is one of these. In contrast to `jax.lax.map` or
    `jax.lax.fori_loop`, it behaves much like `jax.vmap`. In fact, it
    re-implements `in_axes` and `out_axes` and can be used in much the same way
    as `jax.vmap`. However, instead of batching the input, it works through it
    sequentially.

    This implementation makes no claim on being efficient. It explicitly swaps
    around axis in the input and output, potentially allocating more memory
    than strictly necessary and worsening the memory layout.

    For the semantics of `in_axes` and `out_axes` see `jax.vmap`. For the
    semantics of `unroll` see `jax.lax.scan`.
    """
    return partial(_smap, fun, in_axes, out_axes, unroll)


@partial(jax.jit, donate_argnames=("x", ))
def _unsafe_index_update_inplace(x, idx, y):
    return x.at[idx].set(y)


def _lscan(f, init, xs, length=None, unroll=1):
    if unroll != 1:
        raise NotImplementedError()

    if xs is None:
        xs = [None] * length
    carry = init
    ys = None
    lvs = jax.tree_util.tree_leaves(xs)
    # Only use the same device if a single device is used
    lvs_d = [e.devices() for e in lvs if hasattr(e, "devices")]
    like_device = list(set().union(*lvs_d))
    like_device = like_device[0] if len(like_device) == 1 else None
    length = lvs[0].shape[0] if length is None else length
    for i in range(length):
        carry, y = f(carry, tree_map(lambda x: x[i], xs))
        if ys is None:
            # NOTE, `empty_like` will always allocate on the primary device even
            # if `y` is on a different device. Forcefully allocate on the same
            # device as `y`.
            with jax.default_device(like_device):
                ys = tree_map(
                    lambda x: jnp.
                    empty_like(x, shape=(length, ) + jnp.shape(x)), y
                )
        ys = tree_map(
            lambda ys, y: _unsafe_index_update_inplace(ys, i, y), ys, y
        )
    return carry, ys


def lmap(fun, in_axes=0, out_axes=0):
    return partial(_generic_smap, fun, in_axes, out_axes, 1, _scan=_lscan)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2021 Max-Planck-Society
# Author: Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import pytest

from ..common import setup_function, teardown_function


def test_operator_getitem():
    dom = ift.RGSpace([2])
    key = "foo"
    op = ift.ScalingOperator(dom, 2.4)

    op1 = op.ducktape_left(key)
    op2 = op1[key]

    fld = ift.from_random(op1.domain)
    res1 = op1(fld)
    res2 = ift.MultiField.from_dict({key: op2(fld)})
    ift.extra.assert_allclose(res1, res2)
    ift.extra.assert_allclose(res1[key], res2[key])

    lin = ift.Linearization.make_var(fld)
    res1 = op1(lin)
    res2 = op2(lin)
    ift.extra.assert_allclose(res1.val[key], res2.val)

    with pytest.raises(TypeError):
        op[key]



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2020 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest
from numpy.testing import assert_allclose

from ..common import setup_function, teardown_function


def test_ducktape():
    dom = ift.RGSpace(10)
    op = ift.FFTOperator(dom)
    fld = ift.full(dom, 2.)
    lin = ift.Linearization.make_var(fld)
    a = "foo"
    op1 = op.ducktape(a)
    mdom = ift.MultiDomain.make({a: dom})
    assert op1.domain == mdom
    assert op1.target == op.target
    op1 = op.ducktape_left(a)
    assert op1.target == ift.MultiDomain.make({a: op.target})
    assert op1.domain == op.domain
    with pytest.raises(RuntimeError):
        fld.ducktape(a)
    with pytest.raises(RuntimeError):
        lin.ducktape(a)
    fld0 = fld.ducktape_left(a)
    assert fld0.domain == mdom
    lin0 = lin.ducktape_left(a)
    assert lin0.domain == lin.domain
    assert lin0.target == mdom



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from ..domain_tuple import DomainTuple
from ..domains.unstructured_domain import UnstructuredDomain
from ..field import Field
from .linear_operator import LinearOperator


class MaskOperator(LinearOperator):
    """Implementation of a mask response

    Takes a field, applies flags and returns the values of the field in a
    :class:`UnstructuredDomain`.

    Parameters
    ----------
    flags : :class:`nifty8.field.Field`
        Is converted to boolean. Where True, the input field is flagged.
    """
    def __init__(self, flags):
        if not isinstance(flags, Field):
            raise TypeError
        self._domain = DomainTuple.make(flags.domain)
        self._flags = np.logical_not(flags.val)
        self._target = DomainTuple.make(UnstructuredDomain(self._flags.sum()))
        self._capability = self.TIMES | self.ADJOINT_TIMES

        def mask(x):
            return x[self._flags]

        self._jax_expr = mask

    def apply(self, x, mode):
        self._check_input(x, mode)
        x = x.val
        if mode == self.TIMES:
            res = x[self._flags]
            return Field(self.target, res)
        res = np.empty(self.domain.shape, x.dtype)
        res[self._flags] = x
        res[~self._flags] = 0
        return Field(self.domain, res)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2020 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

"""
Some remarks on NIFTy's treatment of random numbers

NIFTy makes use of the `Generator` and `SeedSequence` classes introduced to
`numpy.random` in numpy 1.17.

On first load of the `nifty8.random` module, it creates a stack of
`SeedSequence` objects which contains a single `SeedSequence` with a fixed seed,
and also a stack of `Generator` objects, which contains a single generator
derived from the above seed sequence. Without user intervention, this generator
will be used for all random number generation tasks within NIFTy. This means

- that random numbers drawn by NIFTy will be reproducible across multiple runs
  (assuming there are no complications like MPI-enabled runs with a varying
  number of tasks), and

- that trying to change random seeds via `numpy.random.seed` will have no
  effect on the random numbers drawn by NIFTy.

Users who want to change the random seed for a given run can achieve this
by calling :func:`push_sseq_from_seed()` with a seed of their choice. This will
push a new seed sequence generated from that seed onto the seed sequence stack,
and a generator derived from this seed sequence onto the generator stack.
Since all NIFTy RNG-related calls will use the generator on the top of the stack,
all calls from this point on will use the new generator.
If the user already has a `SeedSequence` object at hand, they can pass this to
NIFTy via :func:`push_sseq`. A new generator derived from this sequence will then
also be pushed onto the generator stack.
These operations can be reverted (and should be, as soon as the new generator is
no longer needed) by a call to :func:`pop_sseq()`.
When users need direct access to the RNG currently in use, they can access it
via the :func:`current_rng` function.


Example for using multiple seed sequences:

Assume that N samples are needed to compute a KL, which are distributed over
a variable number of MPI tasks. In this situation, whenever random numbers
need to be drawn for these samples:

- each MPI task should spawn as many seed sequences as there are samples
  *in total*, using ``sseq = spawn_sseq(N)``

- each task loops over the local samples

  - first pushing the seed sequence for the **global** index of the
    current sample via ``push_sseq(sseq[iglob])```

  - drawing the required random numbers

  - then popping the seed sequence again via ``pop_sseq()``

That way, random numbers should be reproducible and independent of the number
of MPI tasks.

WARNING: do not push/pop the same `SeedSequence` object more than once - this
will lead to repeated random sequences! Whenever you have to push `SeedSequence`
objects, generate new ones via :func:`spawn_sseq()`.
"""

import numpy as np


# Stack of SeedSequence objects. Will always start out with a well-defined
# default. Users can change the "random seed" used by a calculation by pushing
# a different SeedSequence before invoking any other nifty8.random calls
_sseq = [np.random.SeedSequence(42)]
# Stack of random number generators associated with _sseq.
_rng = [np.random.default_rng(_sseq[-1])]


def getState():
    """Returns the full internal state of the module. Intended for pickling.

    Returns
    -------
    state : unspecified
    """
    import pickle
    return pickle.dumps((_sseq, _rng))


def setState(state):
    """Restores the full internal state of the module. Intended for unpickling.


    Parameters
    ----------
    state : unspecified
        Result of an earlier call to `getState`.
    """
    import pickle
    global _sseq, _rng
    _sseq, _rng = pickle.loads(state)


def spawn_sseq(n, parent=None):
    """Returns a list of `n` SeedSequence objects which are children of `parent`

    Parameters
    ----------
    n : int
        number of requested SeedSequence objects
    parent : SeedSequence
        the object from which the returned objects will be derived
        If `None`, the top of the current SeedSequence stack will be used

    Returns
    -------
    list(SeedSequence)
        the requested SeedSequence objects
    """
    if parent is None:
        global _sseq
        parent = _sseq[-1]
    return parent.spawn(n)


def current_rng():
    """Returns the RNG object currently in use by NIFTy

    Returns
    -------
    Generator
        the current Generator object (top of the generatir stack)
    """
    return _rng[-1]


def push_sseq(sseq):
    """Pushes a new SeedSequence object onto the SeedSequence stack.
    This also pushes a new Generator object built from the new SeedSequence
    to the generator stack.

    Parameters
    ----------
    sseq: SeedSequence
        the SeedSequence object to be used from this point

    Notes
    -----
    This function should only be used

    - if you only want to change the random seed once at the very beginning
      of a run, or

    - if the restoring of the previous state has to happen in a different
      Python function. In this case, please make sure that there is a matching
      call to :func:`pop_sseq` for every call to this function!

    In all other situations, it is highly recommended to use the
    :class:`Context` class for managing the RNG state.
    """
    _sseq.append(sseq)
    _rng.append(np.random.default_rng(_sseq[-1]))


def push_sseq_from_seed(seed):
    """Pushes a new SeedSequence object derived from an integer seed onto the
    SeedSequence stack.
    This also pushes a new Generator object built from the new SeedSequence
    to the generator stack.

    Parameters
    ----------
    seed: int
        the seed from which the new SeedSequence will be built

    Notes
    -----
    This function should only be used

    - if you only want to change the random seed once at the very beginning
      of a run, or

    - if the restoring of the previous state has to happen in a different
      Python function. In this case, please make sure that there is a matching
      call to :func:`pop_sseq` for every call to this function!

    In all other situations, it is highly recommended to use the
    :class:`Context` class for managing the RNG state.
    """
    _sseq.append(np.random.SeedSequence(seed))
    _rng.append(np.random.default_rng(_sseq[-1]))


def pop_sseq():
    """Pops the top of the SeedSequence and generator stacks."""
    _sseq.pop()
    _rng.pop()


class Random:
    @staticmethod
    def pm1(dtype, shape):
        if np.issubdtype(dtype, np.complexfloating):
            x = np.array([1+0j, 0+1j, -1+0j, 0-1j], dtype=dtype)
            x = x[_rng[-1].integers(0, 4, size=shape)]
        else:
            x = 2*_rng[-1].integers(0, 2, size=shape)-1
        return x.astype(dtype, copy=False)

    @staticmethod
    def normal(dtype, shape, mean=0., std=1.):
        if not (np.issubdtype(dtype, np.floating) or
                np.issubdtype(dtype, np.complexfloating)):
            raise TypeError("dtype must be float or complex")
        if not np.isscalar(mean) or not np.isscalar(std):
            raise TypeError("mean and std must be scalars")
        if np.issubdtype(type(std), np.complexfloating):
            raise TypeError("std must not be complex")
        if ((not np.issubdtype(dtype, np.complexfloating)) and
                np.issubdtype(type(mean), np.complexfloating)):
            raise TypeError("mean must not be complex for a real result field")
        if np.issubdtype(dtype, np.complexfloating):
            x = np.empty(shape, dtype=dtype)
            x.real = _rng[-1].normal(mean.real, std, shape)
            x.imag = _rng[-1].normal(mean.imag, std, shape)
        else:
            x = _rng[-1].normal(mean, std, shape).astype(dtype, copy=False)
        return x

    @staticmethod
    def uniform(dtype, shape, low=0., high=1.):
        if not np.isscalar(low) or not np.isscalar(high):
            raise TypeError("low and high must be scalars")
        if (np.issubdtype(type(low), np.complexfloating) or
                np.issubdtype(type(high), np.complexfloating)):
            raise TypeError("low and high must not be complex")
        if np.issubdtype(dtype, np.complexfloating):
            x = np.empty(shape, dtype=dtype)
            x.real = _rng[-1].uniform(low, high, shape)
            x.imag = _rng[-1].uniform(low, high, shape)
        elif np.issubdtype(dtype, np.integer):
            if not (np.issubdtype(type(low), np.integer) and
                    np.issubdtype(type(high), np.integer)):
                raise TypeError("low and high must be integer")
            x = _rng[-1].integers(low, high+1, shape)
        else:
            x = _rng[-1].uniform(low, high, shape)
        return x.astype(dtype, copy=False)


class Context:
    """Convenience class for easy management of the RNG state.
    Usage: ::

        with ift.random.Context(seed|sseq):
            code using the new RNG state

    At the end of the scope, the original RNG state will be restored
    automatically.

    Parameters
    ----------
    inp : int or numpy.random.SeedSequence
        The starting information for the new RNG state.
        If it is an integer, a new `SeedSequence` will be generated from it.
    """

    def __init__(self, inp):
        if not isinstance(inp, np.random.SeedSequence):
            inp = np.random.SeedSequence(inp)
        self._sseq = inp

    def __enter__(self):
        self._depth = len(_sseq)
        push_sseq(self._sseq)

    def __exit__(self, exc_type, exc_value, tb):
        pop_sseq()
        if self._depth != len(_sseq):
            raise RuntimeError("inconsistent RNG usage detected")
        return exc_type is None



#!/usr/bin/env python3

import pytest
from numpy.testing import assert_array_equal

import nifty8 as ift

pmp = pytest.mark.parametrize


@pmp("dim,new_shape", (((32,), ((16,),)), ((32, 32), ((16, 16),))))
def test_slice_operator(dim, new_shape):
    domain = ift.RGSpace(dim)
    print(new_shape)
    slice_op = ift.SliceOperator(domain, new_shape, center=False, preserve_dist=True)

    field = ift.Field.from_random(domain=domain, random_type="normal", std=1, mean=0)
    sliced_field = slice_op(field)
    assert sliced_field.val.shape == new_shape[0]
    assert_array_equal(
        field.val[tuple(slice(0, ns) for ns in new_shape[0])], sliced_field.val
    )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest
from numpy.testing import assert_allclose

from ..common import list2fixture, setup_function, teardown_function

pmp = pytest.mark.parametrize
dtype = list2fixture([np.float64, np.float32, np.complex64, np.complex128])
ntries = 10


def test_part_mf_insert():
    dom = ift.RGSpace(3)
    op1 = ift.ScalingOperator(dom, 1.32).ducktape('a').ducktape_left('a1')
    op2 = ift.ScalingOperator(dom, 1).ptw("exp").ducktape('b').ducktape_left('b1')
    op3 = ift.ScalingOperator(dom, 1).ptw("sin").ducktape('c').ducktape_left('c1')
    op4 = ift.ScalingOperator(dom, 1).ducktape('c0').ducktape_left('c')**2
    op5 = ift.ScalingOperator(dom, 1).ptw("tan").ducktape('d0').ducktape_left('d')
    a = op1 + op2 + op3
    b = op4 + op5
    op = a.partial_insert(b)
    fld = ift.from_random(op.domain, 'normal')
    ift.extra.check_operator(op, fld, ntries=ntries)
    ift.myassert(op.domain is ift.MultiDomain.union(
        [op1.domain, op2.domain, op4.domain, op5.domain]))
    ift.myassert(op.target is ift.MultiDomain.union(
        [op1.target, op2.target, op3.target, op5.target]))
    x, y = fld.val, op(fld).val
    assert_allclose(y['a1'], x['a']*1.32)
    assert_allclose(y['b1'], np.exp(x['b']))
    assert_allclose(y['c1'], np.sin(x['c0']**2))
    assert_allclose(y['d'], np.tan(x['d0']))



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_array_equal

from nifty8.re.minisanity import reduced_residual_stats

pmp = pytest.mark.parametrize


@pmp("seed", (33, 42, 43))
def test_reduced_residual_stats_normal_residuals(seed):
    sseq = np.random.SeedSequence(seed)
    rng = np.random.default_rng(sseq)
    ndof = int(1e+3)
    atol, rtol = 1e-14, 1e-14

    normalized_residuals = rng.normal(size=(ndof, ))

    rrs = reduced_residual_stats(normalized_residuals)
    assert_array_equal(rrs.ndof, ndof)
    assert_allclose(
        rrs.mean[0], np.mean(normalized_residuals), atol=atol, rtol=rtol
    )
    assert_allclose(
        rrs.reduced_chisq[0],
        np.sum(normalized_residuals**2) / rrs.ndof,
        atol=atol,
        rtol=rtol
    )


@pmp("seed", (42, 43))
@pmp("complex", (True, False))
def test_reduced_chi_square_dtype(seed, complex):
    sseq = np.random.SeedSequence(seed)
    rng = np.random.default_rng(sseq)
    n_npix = 500
    norm_res = rng.normal(size=(n_npix), )
    if complex:
        norm_res = norm_res + 1j * rng.normal(size=(n_npix), )
    rrs = reduced_residual_stats(norm_res)
    assert rrs.mean.dtype == norm_res.dtype
    # chi^2 should be real and positive
    assert np.abs(rrs.reduced_chisq[0]) == rrs.reduced_chisq[0]

    atol, rtol = 1e-1, 1e-1
    assert_allclose(rrs.reduced_chisq[0], 1., atol=atol, rtol=rtol)
    assert_allclose(rrs.mean, 0., atol=atol, rtol=rtol)


if __name__ == "__main__":
    test_reduced_residual_stats_normal_residuals(33)
    test_reduced_chi_square_dtype(42, True)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
from numpy.testing import assert_allclose, assert_equal

from ..common import list2fixture, setup_function, teardown_function

space = list2fixture([
    ift.RGSpace(4),
    ift.PowerSpace(ift.RGSpace((4, 4), harmonic=True)),
    ift.LMSpace(5),
    ift.HPSpace(4),
    ift.GLSpace(4)
])


def test_property(space):
    diag = ift.Field.from_random(domain=space, random_type='normal')
    D = ift.DiagonalOperator(diag)
    if D.domain[0] != space:
        raise TypeError


def test_times_adjoint(space):
    rand1 = ift.Field.from_random(domain=space, random_type='normal')
    rand2 = ift.Field.from_random(domain=space, random_type='normal')
    diag = ift.Field.from_random(domain=space, random_type='normal')
    D = ift.DiagonalOperator(diag)
    tt1 = rand1.s_vdot(D.times(rand2))
    tt2 = rand2.s_vdot(D.times(rand1))
    assert_allclose(tt1, tt2)


def test_times_inverse(space):
    rand1 = ift.Field.from_random(domain=space, random_type='normal')
    diag = ift.Field.from_random(domain=space, random_type='normal')
    D = ift.DiagonalOperator(diag)
    tt1 = D.times(D.inverse_times(rand1))
    assert_allclose(rand1.val, tt1.val)


def test_times(space):
    rand1 = ift.Field.from_random(domain=space, random_type='normal')
    diag = ift.Field.from_random(domain=space, random_type='normal')
    D = ift.DiagonalOperator(diag)
    tt = D.times(rand1)
    assert_equal(tt.domain[0], space)


def test_adjoint_times(space):
    rand1 = ift.Field.from_random(domain=space, random_type='normal')
    diag = ift.Field.from_random(domain=space, random_type='normal')
    D = ift.DiagonalOperator(diag)
    tt = D.adjoint_times(rand1)
    assert_equal(tt.domain[0], space)


def test_inverse_times(space):
    rand1 = ift.Field.from_random(domain=space, random_type='normal')
    diag = ift.Field.from_random(domain=space, random_type='normal')
    D = ift.DiagonalOperator(diag)
    tt = D.inverse_times(rand1)
    assert_equal(tt.domain[0], space)


def test_adjoint_inverse_times(space):
    rand1 = ift.Field.from_random(domain=space, random_type='normal')
    diag = ift.Field.from_random(domain=space, random_type='normal')
    D = ift.DiagonalOperator(diag)
    tt = D.adjoint_inverse_times(rand1)
    assert_equal(tt.domain[0], space)


def test_diagonal(space):
    diag = ift.Field.from_random(domain=space, random_type='normal')
    D = ift.DiagonalOperator(diag)
    diag_op = D(ift.Field.full(space, 1.))
    assert_allclose(diag.val, diag_op.val)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest
from numpy.testing import assert_allclose

from ..common import list2fixture, setup_function, teardown_function


def _get_rtol(tp):
    if (tp == np.float64) or (tp == np.complex128):
        return 1e-10
    else:
        return 1e-5


pmp = pytest.mark.parametrize
dtype = list2fixture([np.float64, np.float32, np.complex64, np.complex128])
op = list2fixture([ift.HartleyOperator, ift.FFTOperator])


@pmp('d', [0.1, 1, 3.7])
def test_fft1D(d, dtype, op):
    dim1 = 16
    tol = _get_rtol(dtype)
    a = ift.RGSpace(dim1, distances=d)
    b = ift.RGSpace(dim1, distances=1./(dim1*d), harmonic=True)

    fft = op(domain=a, target=b)
    inp = ift.Field.from_random(domain=a, random_type='normal', dtype=dtype, std=7, mean=3)
    out = fft.inverse_times(fft.times(inp))
    assert_allclose(inp.val, out.val, rtol=tol, atol=tol)

    a, b = b, a

    fft = ift.FFTOperator(domain=a, target=b)
    inp = ift.Field.from_random(domain=a, random_type='normal', dtype=dtype, std=7, mean=3)
    out = fft.inverse_times(fft.times(inp))
    assert_allclose(inp.val, out.val, rtol=tol, atol=tol)


@pmp('dim1', [12, 15])
@pmp('dim2', [9, 12])
@pmp('d1', [0.1, 1, 3.7])
@pmp('d2', [0.4, 1, 2.7])
@pmp('nthreads', [-1, 1, 2, 3, 4])
def test_fft2D(dim1, dim2, d1, d2, dtype, op, nthreads):
    ift.set_nthreads(nthreads)
    ift.myassert(ift.nthreads() == nthreads)
    tol = _get_rtol(dtype)
    a = ift.RGSpace([dim1, dim2], distances=[d1, d2])
    b = ift.RGSpace(
        [dim1, dim2], distances=[1./(dim1*d1), 1./(dim2*d2)], harmonic=True)

    fft = op(domain=a, target=b)
    inp = ift.Field.from_random(domain=a, random_type='normal', dtype=dtype, std=7, mean=3)
    out = fft.inverse_times(fft.times(inp))
    assert_allclose(inp.val, out.val, rtol=tol, atol=tol)

    a, b = b, a

    fft = ift.FFTOperator(domain=a, target=b)
    inp = ift.Field.from_random(domain=a, random_type='normal', dtype=dtype, std=7, mean=3)
    out = fft.inverse_times(fft.times(inp))
    assert_allclose(inp.val, out.val, rtol=tol, atol=tol)
    ift.set_nthreads(1)


@pmp('index', [0, 1, 2])
def test_composed_fft(index, dtype, op):
    tol = _get_rtol(dtype)
    a = [a1, a2,
         a3] = [ift.RGSpace((32,)),
                ift.RGSpace((4, 4)),
                ift.RGSpace((5, 6))]
    fft = op(domain=a, space=index)

    inp = ift.Field.from_random(domain=(a1, a2, a3), random_type='normal', dtype=dtype, std=7, mean=3)
    out = fft.inverse_times(fft.times(inp))
    assert_allclose(inp.val, out.val, rtol=tol, atol=tol)


@pmp('space', [
    ift.RGSpace(128, distances=3.76, harmonic=True),
    ift.RGSpace((15, 27), distances=(.7, .33), harmonic=True),
    ift.RGSpace(73, distances=0.5643)
])
def test_normalisation(space, dtype, op):
    tol = 10*_get_rtol(dtype)
    cospace = space.get_default_codomain()
    fft = op(space, cospace)
    inp = ift.Field.from_random(domain=space, random_type='normal', dtype=dtype, std=1, mean=2)
    out = fft.times(inp)
    fft2 = op(cospace, space)
    out2 = fft2.inverse_times(inp)
    zero_idx = tuple([0]*len(space.shape))
    assert_allclose(
        inp.val[zero_idx], out.s_integrate(), rtol=tol, atol=tol)
    assert_allclose(out.val, out2.val, rtol=tol, atol=tol)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2019-2021 Max-Planck-Society
# Copyright(C) 2022 Max-Planck-Society, Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest

from ..common import setup_function, teardown_function

pmp = pytest.mark.parametrize


def _l2error(a, b):
    return np.sqrt(np.sum(np.abs(a-b)**2)/np.sum(np.abs(a)**2))


@pmp('eps', [1e-2, 1e-4, 1e-7, 1e-10, 1e-11, 1e-12, 2e-13])
@pmp('nxdirty', [32, 128])
@pmp('nydirty', [32, 48, 128])
@pmp('N', [1, 10, 100])
def test_gridding(nxdirty, nydirty, N, eps):
    uv = ift.random.current_rng().random((N, 2)) - 0.5
    vis = (ift.random.current_rng().standard_normal(N)
           + 1j*ift.random.current_rng().standard_normal(N))

    if N > 2:
        uv[-1] = 0
        uv[-2] = 1e-5
    # Nifty
    dom = ift.RGSpace((nxdirty, nydirty), distances=(0.2, 1.12))
    dstx, dsty = dom.distances
    uv[:, 0] = uv[:, 0]/dstx
    uv[:, 1] = uv[:, 1]/dsty
    Op = ift.Gridder(dom, uv=uv, eps=eps)
    vis2 = ift.makeField(ift.UnstructuredDomain(vis.shape), vis)

    pynu = Op(vis2).val
    # DFT
    x, y = np.meshgrid(
        *[-ss/2 + np.arange(ss) for ss in [nxdirty, nydirty]], indexing='ij')
    dft = pynu*0.
    for i in range(N):
        dft += (
            vis[i]*np.exp(2j*np.pi*(x*uv[i, 0]*dstx + y*uv[i, 1]*dsty))).real
    ift.myassert(_l2error(dft, pynu) < eps)


def test_cartesian():
    nx, ny = 32, 42
    dstx, dsty = 0.3, 0.2
    dom = ift.RGSpace((nx, ny), (dstx, dsty))

    kx = np.fft.fftfreq(nx, dstx)
    ky = np.fft.fftfreq(ny, dsty)
    uu, vv = np.meshgrid(kx, ky)
    tmp = np.vstack([uu[None, :], vv[None, :]])
    uv = np.transpose(tmp, (2, 1, 0)).reshape(-1, 2)

    op = ift.Gridder(dom, uv=uv).adjoint

    fld = ift.from_random(dom, 'normal')
    arr = fld.val

    fld2 = ift.makeField(dom, np.roll(arr, (nx//2, ny//2), axis=(0, 1)))
    res = op(fld2).val.reshape(nx, ny)

    fft = ift.FFTOperator(dom.get_default_codomain(), target=dom).adjoint
    vol = ift.full(dom, 1.).s_integrate()
    res1 = fft(fld).val

    np.testing.assert_allclose(res, res1*vol)


@pmp('eps', [1e-2, 1e-6, 2e-13])
@pmp('nxdirty', [32, 128])
@pmp('nydirty', [32, 48, 128])
@pmp('N', [1, 10, 100])
def test_build(nxdirty, nydirty, N, eps):
    dom = ift.RGSpace([nxdirty, nydirty])
    uv = ift.random.current_rng().random((N, 2)) - 0.5
    RF = ift.Gridder(dom, uv=uv, eps=eps)

    # Consistency checks
    flt = np.float64
    cmplx = np.complex128
    # We set rtol=eps here, because the gridder operator only guarantees
    # adjointness to this accuracy.
    ift.extra.check_linear_operator(RF, cmplx, flt, only_r_linear=True, rtol=eps)


@pmp('eps', [1e-2, 1e-4, 1e-7, 1e-10, 1e-11, 1e-12, 2e-13])
@pmp('nxdirty', [32, 128])
@pmp('N', [1, 10, 100])
def test_nu1d(nxdirty, N, eps):
    pos = ift.random.current_rng().random((N)) - 0.5
    vis = (ift.random.current_rng().standard_normal(N)
           + 1j*ift.random.current_rng().standard_normal(N))

    if N > 2:
        pos[-1] = 0
        pos[-2] = 1e-5
    # Nifty
    dom = ift.RGSpace((nxdirty), distances=0.2)
    dstx = dom.distances
    pos = pos / dstx
    Op = ift.Nufft(dom, pos=pos[:, None], eps=eps)
    vis2 = ift.makeField(ift.UnstructuredDomain(vis.shape), vis)
    pynu = Op(vis2).val
    # DFT
    x = -nxdirty/2 + np.arange(nxdirty)

    dft = pynu*0
    for i in range(N):
        dft += (vis[i]*np.exp(2j*np.pi*(x*pos[i]*dstx))).real
    ift.myassert(_l2error(dft, pynu) < eps*10)


@pmp('eps', [1e-2, 1e-4, 1e-7, 1e-10, 1e-11, 1e-12, 2e-13])
@pmp('nxdirty', [32, 128])
@pmp('nydirty', [32, 48, 128])
@pmp('N', [1, 10, 100])
def test_nu2d(nxdirty, nydirty, N, eps):
    uv = ift.random.current_rng().random((N, 2)) - 0.5
    vis = (ift.random.current_rng().standard_normal(N)
           + 1j*ift.random.current_rng().standard_normal(N))

    if N > 2:
        uv[-1] = 0
        uv[-2] = 1e-5
    # Nifty
    dom = ift.RGSpace((nxdirty, nydirty), distances=(0.2, 1.12))
    dstx, dsty = dom.distances
    uv[:, 0] = uv[:, 0]/dstx
    uv[:, 1] = uv[:, 1]/dsty
    Op = ift.Nufft(dom, pos=uv, eps=eps)
    vis2 = ift.makeField(ift.UnstructuredDomain(vis.shape), vis)

    pynu = Op(vis2).val
    # DFT
    x, y = np.meshgrid(
        *[-ss/2 + np.arange(ss) for ss in [nxdirty, nydirty]], indexing='ij')
    dft = pynu*0.
    for i in range(N):
        dft += (
            vis[i]*np.exp(2j*np.pi*(x*uv[i, 0]*dstx + y*uv[i, 1]*dsty))).real
    ift.myassert(_l2error(dft, pynu) < eps*10)


@pmp('eps', [1e-2, 1e-4, 1e-7, 1e-11])
@pmp('nxdirty', [32, 128])
@pmp('nydirty', [32, 48])
@pmp('nzdirty', [32, 54])
@pmp('N', [1, 10])
def test_nu3d(nxdirty, nydirty, nzdirty, N, eps):
    pos = ift.random.current_rng().random((N, 3)) - 0.5
    vis = (ift.random.current_rng().standard_normal(N)
           + 1j*ift.random.current_rng().standard_normal(N))
    # Nifty
    dom = ift.RGSpace((nxdirty, nydirty, nzdirty), distances=(0.2, 1.12, 0.7))
    dstx, dsty, dstz = dom.distances
    pos[:, 0] = pos[:, 0]/dstx
    pos[:, 1] = pos[:, 1]/dsty
    pos[:, 2] = pos[:, 2]/dstz
    Op = ift.Nufft(dom, pos=pos, eps=eps)
    vis2 = ift.makeField(ift.UnstructuredDomain(vis.shape), vis)

    pynu = Op(vis2).val
    # DFT
    x, y, z = np.meshgrid(
        *[-ss/2 + np.arange(ss) for ss in [nxdirty, nydirty, nzdirty]], indexing='ij')
    dft = pynu*0.
    for i in range(N):
        dft += (
            vis[i]*np.exp(2j*np.pi*(x*pos[i, 0]*dstx + y*pos[i, 1]*dsty + z*pos[i, 2]*dstz))).real
    ift.myassert(_l2error(dft, pynu) < eps*10)


@pmp('eps', [1e-2, 1e-6, 3e-13])
@pmp('space', [ift.RGSpace(128),
               ift.RGSpace([32, 64]),
               ift.RGSpace([10, 27, 32])])
@pmp('N', [1, 10, 100])
def test_build_nufft(space, N, eps):
    pos = ift.random.current_rng().random((N, len(space.shape))) - 0.5
    RF = ift.Nufft(space, pos=pos, eps=eps)
    flt = np.float64
    cmplx = np.complex128
    # We set rtol=eps here, because the gridder operator only guarantees
    # adjointness to this accuracy.
    ift.extra.check_linear_operator(RF, cmplx, flt, only_r_linear=True, rtol=eps)
